#!/usr/bin/env python
#Boa:PyApp:main
import math
modules ={}

def diretto (F,dmin,SH):
    '''ricavando diametro all' ingersso da diam allintersezione'''
    th=math.atan2(float(dmin/2),float(F))/4
    dmax=2*SH*math.tan(th)+dmin
    return dmax

def inverso(F,dmax,H):
    '''ricava il dianetri allintersezione dal diametro d'ingresso
    con metodo ricorsico.  stessa unita' di misura.'''
    rIn=float(dmax)/2

    def ThSh(R):return math.atan2(float(R),float(F))/4
    def Rint(rIn,th0):return rIn-H*math.tan(th) #trova raggio allintersezione da angolo e rIn
    r=rIn
    rOld=0
    while (abs(r-rOld)>10e-8):
        rOld=r
        th=ThSh(r)        
        r=Rint(rIn,th)
    return r*2   

def daa(y0,angolo):
    '''dato un punto di partenza (0,y0), e l'angolo con l'asse y 
    per una retta con coefficiente angolare negativo, 
    restituisce la x nel di ordinata y1'''
    return y0*math.tan(angolo)

def angsep(F,H,r,theta):
    alfa=math.atan2(r,F)/4
    H1=H2=H
    #trova coordinata minima pupilla uscita, il raggio colpisce lintersezione p-h
    rout=r-daa(H2,alfa+theta)
    #trova prossimo raggio corrispondente
    def ThSh(R):return math.atan2(float(R),float(F))/4
    def Rint(rint,th0,H):return rint-H*math.tan(th) #trova raggio all'uscita
    rOld=0
    while (abs(rout-rOld)>10e-8):
        rOld=rout
        th=ThSh(rout)        
        rout=Rint(rout,3*th,H2)
    #r1 e' il raggio all'uscita
    rint= rout+H2*math.tan(3*th) 
    rin= rint+H1*math.tan(th)
    #calcola angsep
    angsep=math.atan2(r-rin,H1)
    return angsep

def main(F,H,r_vec,theta):
    '''assumendo shell con altezza uguale per parabola e iperbole,
    calcola la minima separazione
    angolare fra le shell per evitare ostruzione all'uscita per i
    raggi del vettore r_vec per l'angolo fuori asse theta'''
    sep_vec=[angsep(F,H,r,theta) for r in r_vec]
    print(r_vec, sep_vec)
    

if __name__ == '__main__':
    rmin,rmax=175,350
    H=300
    F=20000
    npunti=100
    r_vec=[rmin+xx*(rmax-rmin)/(npunti+1) for xx in range(npunti)]
    theta=1/(60*57.) #1 primo
    main(F,H,r_vec,theta)



import numpy as np


def any_is_none(val):
    """Return True if any of the elements in an iterable is none.

    Needed because builtins.any fails on non iterable and np.any gives
      inconsistent results (signaled to numpy).
    Note, it is not recursive (if nested iterable,
    only first level is checked)."""
    res = val is None if np.ndim(val) == 0 else np.any([vv is None for vv in val])
    return res


if __name__ == "__main__":
    tv = [None, [None], [None, 2], 'cane']  # test values
    for v in tv: print(any_is_none(v))



"""here coating stress fit and optimization functions from C1S05_stress_compensation_target.ipynb"""


import numpy as np 
import matplotlib.pyplot as plt
import os 
import pandas as pd
from plotting.captions import legendbox
from pySurf.data2D import get_stats
from dataIO.outliers import remove_outliers
from pyProfile.profile import PSF_spizzichino,line
from pySurf.data2D import plot_data, crop_data, level_data
from plotting.multiplots import diff_images
from dataIO.span import span
from dataIO.fn_add_subfix import fn_add_subfix
from utilities.imaging import fitting as fit
from plotting.backends import maximize
from dataIO.any_is_none import any_is_none


def fom_rms(tdata,x=None,y=None):
    """Fit surface after plane level"""
    #tdata=tdata[ci0:ci1,ci0:ci1]
    #tdata=tdata-fit.legendre2d(tdata,1,1)[0]
    return np.nanstd(level_data(tdata)[0])

def fom_sl(tdata,x=None,y=None):
    """Fit slope of surface after plane level"""
    """This was taking a second argument dy, changed 2019/03/28 to data,x,y,
    do a workaround to keep old compatibility."""
    #tdata=tdata[ci0:ci1,ci0:ci1]
    
    if y is None and x is not None:
        dy=x
    else:    
        dy=np.diff(diff.y)[0]
    tdata=tdata-fit.legendre2d(tdata,1,1)[0]
    grad=np.gradient(tdata)
    slopeax=grad[0][:-1,:]/1000.*206265/dy  #in arcseconds from um
    return np.nanstd(slopeax)

def comp2(d1,d2,roi=None):
    #adapted (partially) to data2d
    """make 4 panel plot with data, differences and difference slope for two data sets on same x and y.
    Return four axis for plot customization.
    all data and all stats are leveled.
    
    roi: if passed, plot a corresponding rectangle."""
    partial=False  #not implemented, if a roi is defined, adds a legend with partial statistics
    if roi is not None:
        rect=np.array([(roi[0][0],roi[1][0]),
                       (roi[0][0],roi[1][1]),
                       (roi[0][1],roi[1][1]),
                       (roi[0][1],roi[1][0]),
                       (roi[0][0],roi[1][0])])   
    
    plt.figure(6)
    plt.clf()
    maximize()
    
    import pdb
    #pdb.set_trace()
    diff=d1-d2
    ax1=plt.subplot(141)
    d1.level().plot(title='Data')
    plt.clim(*remove_outliers(d1.level().data,nsigma=2,itmax=3,span=True))
    if roi is not None:
        plt.plot(rect[:,0], rect[:,1], '--',c='black', lw=2)
        legendbox(get_stats(d1.crop(*roi).level()()[0]),loc=4)
        
    ax2=plt.subplot(142,sharex=ax1,sharey=ax1)
    d2.level().plot(title='Simulation')
    
    plt.clim(*remove_outliers(d2.level().data,nsigma=2,itmax=3,span=True))
    if roi is not None:
        plt.plot(rect[:,0], rect[:,1], '--',c='black', lw=2)
        legendbox(get_stats(d2.crop(*roi).level()()[0]),loc=4)
        
    ax3=plt.subplot(143,sharex=ax1,sharey=ax1)
    diff.level().plot(title='Difference')    
    plt.clim(*remove_outliers(diff.level().data,nsigma=2,itmax=3,span=True))
    if roi is not None:
        plt.plot(rect[:,0], rect[:,1], '--',c='black', lw=2)
        legendbox(get_stats(diff.crop(*roi).level()()[0]),loc=4)
    
    ax4=plt.subplot(144,sharex=ax1,sharey=ax1)
    
    #2019/03/28 replaced with object
    #grad=np.gradient(diff.level().data)
    #dy=np.diff(diff.y)[0]
    #slopeax=grad[0]/1000.*206265/dy
    #x,y=diff.x,diff.y
    #plot_data(slopeax,x,y,units=['mm','mm','arcsec'],title='Axial Slope',stats=True) #,
              #vmin=-250,vmax=250)
    #plt.clim(*remove_outliers(slopeax,nsigma=2,itmax=3,span=True))
    #if roi is not None:
    #    plt.plot(rect[:,0], rect[:,1], '--',c='black', lw=2)
    #    legendbox(get_stats(*crop_data(slopeax,x,y,*roi),units=['mm','mm','arcsec']),loc=4)
    
    slopeax=diff.slope(scale=(1.,1.,1000.))[1]
    slopeax.name='Axial Slope'
    slopeax.plot()
    if roi is not None:
        plt.plot(rect[:,0], rect[:,1], '--',c='black', lw=2)
        legendbox(get_stats(slopeax.crop(*roi).level()()[0]),loc=4)   
    
    axes = [ax1,ax2,ax3,ax4]
    '''
    if roi is not None:  #update legend to include total and partial stats.
        for ax in axes:
            for art in ax.artists:
                if isinstance(art,matplotlib.legend.Legend):
                    p = art.get_patches()
                    t = art.get_texts()
                    newleg=Legend(ax,p,t,loc=2,handletextpad=0, handlelength=0)
                    ax.add_artist(newleg)
                    art.remove()
    '''
    return axes

def simple_regressor(d1,d2):
    """Use analytical for simple regression formula to determine best fit scale factor.
       Data are leveled before calculation of fom, but note that leveling can be altered by nan. 
       The function is defined on 2D data (not on Data2D objects), d1 and d2 must have same shape."""
    import pdb
    #pdb.set_trace()
    d2[np.isnan(d1)]=np.nan #this is to account in leveling for points that are nan in difference (and product, but non in d2)
    tbest=np.nansum(level_data(d1)[0]*level_data(d2)[0])/np.nansum(level_data(d2)[0]**2)
    return tbest
    
def scale_fit2(d1,d2,testval=None,fom=fom_rms,outfile='',roi=None, crop=False, dis= True):
    """d1 and d2 are Data2D objects. Returns value for scaling tbest of d2
    that minimizes fom(d1-d2*tbest). e.g. fom = rms of surface/slope.
    Plot rms as a function of scale factor (if testval are passed for raster scan) 
        and 4 panel plot of best result (use comp2) either cropped (if crop is True) 
        or full.
    
    if testval is set to None, best fit is determined by analytical formula (minimizing
        rms of surface difference). If tesval is array, all values are tested. 
    
    if `roi` is passed, fom is calculated on the roi only.
    `crop` flag controls the plotting. If True, only the roi is plotted and 
    color scale is adjusted accordingly. Otherwise, entire region is plotted and
    a rectangle around the roi is plotted.
    dis=True print information and display plots in notebook.
    
    all fom and stats are calculated on leveled data, ideally I want to have stats in 
    roi unleveled, but it is safer this way for now.
    """
    
    #dy=np.diff(d1.y)[0]
    d2=d2.resample(d1)
    
    res=[]
    if roi is None: 
        diff=d1-d2
        roi = [span(diff.x),span(diff.y)]  #stupid way of doing it
    
    if testval is None:
        #use analytical formula
        tbest=simple_regressor(d1.crop(*roi).data,d2.crop(*roi).data)
        dtest=(d1-(d2*tbest)).crop(*roi).level()
        rbest=fom(*dtest())
    else:
        #test all values
        for i,t in enumerate(testval):  
            #test each testval, create a plot (on file only). If dis is True, print vals
            diff=d1-d2*t 
            f=fom(*(diff.crop(*roi))())
            res.append(f)
            axes=comp2(d1,d2*t,roi=roi)
            plt.tight_layout(rect=[0, 0.03, 1, 0.95])
            plt.suptitle('factor: %s'%t)
            if outfile:
                plt.savefig(fn_add_subfix(outfile,'_%03i'%(i+1),'.jpg'))
            if dis: print (i,t,f)        
        #select best value
        res=np.array(res)
        ibest=np.argmin(res)   
        tbest=testval[np.argmin(res)]
        rbest=res[ibest]
        #make plot with rms as a function of testval
        plt.figure(5)
        plt.clf()
        plt.plot(testval,res)
        plt.title('best factor:%6.3f, FOM surf:%6.3f'%(tbest,res[ibest]))
        plt.grid()
        if outfile:
            plt.savefig((fn_add_subfix(outfile,'','.jpg',pre='FOMtrend_')))
        if dis: display(plt.gcf())
    
    #plot and display (if dis) best fit data (cropped if crop is selected, full otherwise).
    if crop:  
        d1 = d1.crop(*roi).level()
        d2 = d2.crop(*roi).level()
    plt.figure()
    a=comp2(d1,d2*tbest,roi=roi)
    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
    plt.suptitle('best factor:%6.3f, FOM surf:%6.3f'%(tbest,rbest))
    if dis: display(plt.gcf())
    if outfile: plt.savefig(fn_add_subfix(outfile,'_fit','.jpg'))
    
    return tbest



def trioptimize(d1,d2,fom,testvals=None,rois=None,outfile=None,noplot=False,dis=True): 
    """run optimizations using scale_fit2 over a set of cases
    defined by the values in matching lists of testvals (array of scaling test values) 
    and rois ([[xmin,xmax],[ymin,ymax]]). 
    
    Returns a pandas data structure with parameters and best scaling values for each case,
        with fields ['tbest','testval','roi','fom_roi'].
    
    Plot 
 
    d1 and d2 are Data2D objects.
    testvals: list of vectors to use as test values for each of the ROI
    rois:    list of rois (same len as testvals), in form [[xmin,xmax],[ymin,ymax]], 
        set to None to use for range.
    outfile:
    noplot:
    dis:
    """
    
    if outfile is not None:
        outname = os.path.basename (outfile)
        outfolder = os.path.dirname (outfile)
    else:
        outname = None
    
    plt.close('all')
    res=pd.DataFrame(index=['tbest','testval','roi','fom_roi','fom_initial']).transpose()

    if any_is_none(testvals): 
        if fom != fom_rms:
            raise ValueError('analytical fit (without testval) supported only for fom_rms')
        if testvals is None: testvals=[None]*len(rois)
    
    #Minimizes on each subaperture
    for i,(roi,testval) in enumerate(zip(rois,testvals)):  
        name = 'crop%02i'%(i+1) if outname is None else outname+'_crop%02i'%(i+1)
        #if dis print optimization parameters
        if dis:
            print ("\n\n\n===   CASE %i: %s  ===\n"%(i+1,name))
            if roi is None:
                print ('FULL APERTURE')
            else:
                print ("CROP: [%.2f:%.2f] [%.2f:%.2f]"%tuple(roi[0]+roi[1]))
            print ("TEST RANGE:",span(testval))
            dtest=(d1-d2).crop(*roi) if roi is not None else d1-d2
            fom0=fom(*dtest())
            print ("INITIAL FOM:",fom0)
        
        #do optimization, plot bestfit on ROI, displaying plot if dis is True
        s=scale_fit2(d1,d2,testval, dis=dis if roi is not None else False,  #it will plot later if no roi 
                  fom=fom,outfile='' if outname is None else fn_add_subfix (outfile,
                  '_crop%02i'%(i+1)),roi=roi,crop=dis) #if dis is set, plot roi, otherwise skip

        #calculate and store best fit
        ftest =fom(*((d1-d2*s).crop(*roi))()) if roi is not None else fom(*(d1-d2*s)())
        b=pd.Series([s,testval,roi,ftest,fom0],index=['tbest','testval','roi','fom_roi','fom_initial'], #fom_initial added 04/02
            name=name)
        res = res.append(b)
        
        #plot best fit case full range unless noplot
        if not noplot:
            plt.close('all')
            comp2(d1,d2*s,roi=roi) # plots entire area with rectangle
            dtest=(d1-d2*s)
            plt.tight_layout(rect=[0, 0.03, 1, 0.95])
            plt.suptitle('best factor:%6.3f, FOM surf:%6.3f'%(s,fom(*dtest())))
            if outname:
                plt.savefig(fn_add_subfix(outfile,'_croproi_%02i'%(i+1),'.png'))
            if dis: display(plt.gcf())
        if dis: print ("BEST SCALE: %.3f"%s," FOM:",ftest)
        
    if dis: print ("\n===\nSUMMARY:")
    #a=pd.DataFrame(index=['tbest','testval','roi','fom_roi']).transpose()
    for roi,r in zip(rois,res.iterrows()):
        if dis:
            if roi is None:
                print ('FULL APERTURE--> BEST SCALE: ', r[1]['tbest'])
            else:
                #pdb.set_trace()
                #not sure why this is needed, but r[0] is the dataseries name, r[1] the dataseries
                print ("CROP: [%.2f:%.2f] [%.2f:%.2f]"%tuple(roi[0]+roi[1]),"--> BEST SCALE: ",r[1]['tbest'])
    print ("=======================")
    
    return res
    
def reoptimize(scales, step=None, nstep=5):
    """run a numerical optimization starting from a centered value and a number or steps and stepsize.
    Optimization parameters are taken from a pandas dataFrame that has fields 
        It can be used to numerically rerun an analytical fit (or to refine a previous fit).
        2019/03/28 quickly taken from notebook. """
    
    #initialize allres
    #step=0.001
    #nsteps=5 

    print(scales['tbest','fom_roi','roi'])
    tv=scales['tbest'].values
    rois=[None if np.isnan(s).all() else s for s in scales['roi'].values] #pd makes them nan when they were None, convert back

    scales2=trioptimize(sdiff,df,fom=fom,
               testvals=[np.arange(x0,y0,step) for x0,y0 in zip(tv-step*5,tv+step*5)],
               rois=rois,dis =True,
               outfile=os.path.join(outfolder,prfolder,'VM_stress_fit'))
    print(scales2['tbest','fom_roi','roi'])
    return scales2


import os
import pdb
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
from dataIO.fn_add_subfix import fn_add_subfix
from dataIO.functions import update_docstring
from dataIO import outliers  #necessary to avoid conflict 
from dataIO.span import span
from plotting.add_clickable_markers import add_clickable_markers2

from pySurf import data2D, points
from pySurf.affine2D import find_affine
from pySurf.data2D import apply_transform as apply_transform_data
from pySurf.data2D import (crop_data, data_from_txt, data_histostats,
                           level_data, plot_data, projection, read_data,
                           register_data, resample_data, rotate_data,
                           save_data, slope_2D, subtract_data, sum_data,
                           transpose_data, get_stats)
from pySurf.points import matrix_to_points2, points_autoresample
from pySurf.psd2d import (plot_psd2d, plot_rms_power, psd2d, psd2d_analysis,
                          rms_power)
from pySurf.readers.format_reader import auto_reader

"""
2018/06/07 v1.3
v1.2 was not convenient, switch back to same interface as 1.1:
don't modify self, return copy.

2018/06/06 v 1.2
After attempt to modify interface in a way that modify self, 
switch back to same interface as 1.1. It is inconvenient to have 
always to create a copy, in particular in interactive mode when same 
operation can be repeated multiple times you need to inizialize data 
every time (e.g. rotate).

2018/06/06 v1.1
methods are written to consistently return a copy without modifying self.
But this doesn't work, always need to assign, cannot chain (with property 
assignment, it works on methods) or apply to set of data in list.

Give some attention to inplace operators that can link to external data. 
At the moment when class is created from data, x, y these are assigned directly 
to the property resulting in a link to the original data. Some methods have 
inplace operations that will reflect on initial data, others don't."""

"""
programming notes:
-inplace methods-
A method acting on Data2D (e.g. level) can modify (o reassign to) self and 
return self or return a copy.

e.g. a method crop (or remove nancols) that return slices of data and x/y 
reassign to self.data/x/y views of original data.

A method that reassigns self.data=...
most of the time stays linked to original data.
If inplace changes (e.g. element-wise assignment) are then performed on the 
property, the change is reflected in the original data.
See notebook Python Programming Notes.

To remove this behavior, a copy of the data or view must be performed at some 
point (e.g. explicitly self.data.copy() or implicitly deepcopy(self).



-attempt to subclass ndarray-
def __new__(subtype, shape, dtype=float, buffer=None, offset=0,
            strides=None, order=None, info=None):
    # Create the ndarray instance of our type, given the usual
    # ndarray input arguments.  This will call the standard
    # ndarray constructor, but return an object of our type.
    # It also triggers a call to InfoArray.__array_finalize__
    obj = super(Data2D, subtype).__new__(subtype, shape, dtype,    #(InfoArray
                                            buffer, offset, strides,
                                            order)
    # set the new 'info' attribute to the value passed
    #obj.info = info
    # Finally, we must return the newly created object:
    return obj

def __array_finalize__(self, obj):
    # ``self`` is a new object resulting from
    # ndarray.__new__(InfoArray, ...), therefore it only has
    # attributes that the ndarray.__new__ constructor gave it -
    # i.e. those of a standard ndarray.
    #
    # We could have got to the ndarray.__new__ call in 3 ways:
    # From an explicit constructor - e.g. InfoArray():
    #    obj is None
    #    (we're in the middle of the InfoArray.__new__
    #    constructor, and self.info will be set when we return to
    #    InfoArray.__new__)
    if obj is None: return
    # From view casting - e.g arr.view(InfoArray):
    #    obj is arr
    #    (type(obj) can be InfoArray)
    # From new-from-template - e.g infoarr[:3]
    #    type(obj) is InfoArray
    #
    # Note that it is here, rather than in the __new__ method,
    # that we set the default value for 'info', because this
    # method sees all creation of default objects - with the
    # InfoArray.__new__ constructor, but also with
    # arr.view(InfoArray).
    #self.info = getattr(obj, 'info', None)
    # We do not need to return anything
"""


class Data2D(object):  # np.ndarrays
    """A class containing 2D data with x and y coordinates and methods for analysis.

    Args:
        object ([type]): [description]

    Raises:
        ValueError: [description]
        ValueError: [description]
        ValueError: [description]
        ValueError: [description]
        ValueError: [description]
        ValueError: [description]

    Returns:
        [type]: [description]
    """    

    """
    
    def __new__(subtype, shape, dtype=float, buffer=None, offset=0,
                strides=None, order=None, info=None):
        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments.  This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to InfoArray.__array_finalize__
        obj = super(Data2D, subtype).__new__(subtype, shape, dtype,    
                                            #(InfoArray
                                                buffer, offset, strides,
                                                order)
        # set the new 'info' attribute to the value passed
        #obj.info = info
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(InfoArray, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. InfoArray():
        #    obj is None
        #    (we're in the middle of the InfoArray.__new__
        #    constructor, and self.info will be set when we return to
        #    InfoArray.__new__)
        if obj is None: return
        # From view casting - e.g arr.view(InfoArray):
        #    obj is arr
        #    (type(obj) can be InfoArray)
        # From new-from-template - e.g infoarr[:3]
        #    type(obj) is InfoArray
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'info', because this
        # method sees all creation of default objects - with the
        # InfoArray.__new__ constructor, but also with
        # arr.view(InfoArray).
        #self.info = getattr(obj, 'info', None)
        # We do not need to return anything
    """
    #def __repr__(self):
        #breakpoint()
        #pass
        #repr(self)
        #return self

    def __init__(
        self,
        data=None,
        x=None,
        y=None,
        file=None,
        reader=None,
        units=None,
        name=None,
        *args,
        **kwargs
    ):
        """
        A class for 2D data with coordinates and their analysis.
        
        Can be initialized with data | data, x, y | file | file, x, y.
        if x and y are coordinates if match number of elements, 
            or used as range if two element. 
        If provided together with file, they override x and y 
            read from file.
        
        Function methods return a copy with new values and don't alter original
           object. Reassign to variable to use as modifier:
           e.g. a=a.level( ... )

        Args:
            data (2D array or string): 2D data or file name (suitable reader must 
            provided).
            
            x, y (array): coordinates or ranges.
            
            file (str): alternative way to provide a data file.
            
            units (str array): 3-element array with units symbols for `x`, `y`, `data`  
            
            reader (function): reader function (see `pySurf.readers.instrumentReader`).
            
            name (str): sets the name of created object.  

            *args, **kwargs: optional arguments for `pySurf.data2D.register_data`. 
        """

        # from pySurf.readers.instrumentReader import reader_dic

        # pdb.set_trace()

        if isinstance(data, str):
            print("first argument is string, use it as filename")
            file = data
            data = None
        # pdb.set_trace()
        self.file = file  # initialized to None if not provided
        if file is not None:
            assert data is None
            # store in xrange values for x if were passed
            xrange = span(x) if x is not None else None
            yrange = span(y) if y is not None else None

            # pdb.set_trace()
            if reader is None:
                reader = auto_reader(file)  # returns a reader
            # calling without arguments skips register, however skips also reader argumnets, temporarily arranged with pop in read_data to strip all arguments for
            data, x, y = read_data(file, reader, *args, **kwargs)
            # register data and pass the rest to reader
            # pdb.set_trace()

            if np.size(x) == data.shape[1]:
                self.x = x
            elif np.size(x) == 2:
                # y and yrange are identical
                print("WARNING: 2 element array provided for X, uses as range.")
                x = np.linspace(*xrange, data.shape[1])
            elif xrange is not None:
                print(
                    "wrong number of elements for x (must be 2 or xsize_data [%i]), it is instead %i"
                    % (np.size(data)[1], np.size(x))
                )
                raise ValueError

            if np.size(y) == data.shape[0]:
                self.y = y
            elif np.size(y) == 2:
                # y and yrange are identical
                print("WARNING: 2 element array provided for Y, uses as range.")
                x = np.linspace(*yrange, data.shape[0])
            elif yrange is not None:
                print(
                    "wrong number of elements for y (must be 2 or ysize_data [%i]), it is instead %i"
                    % (np.size(data)[0], np.size(y))
                )
                raise ValueError

            # set self.header to file header if implemented in reader, otherwise set to empty string""
            try:
                # kwargs['header']=True
                # self.header=reader(file,header=True,*args,**kwargs)
                self.header = reader(file, header=True, *args, **kwargs)
            except TypeError:  # unexpected keyword if header is not implemented
                self.header = ""
                # raise
        else:
            if data is not None:
                if len(data.shape) != 2:
                    # pdb.set_trace()
                    print(
                        "WARNING: data are not bidimensional, results can be unpredictable!"
                    )
                if x is None:
                    x = np.arange(data.shape[1])
                if y is None:
                    y = np.arange(data.shape[0])

                # if data is not None:
                # se read_data calls register, this
                data, x, y = register_data(data, x, y, *args, **kwargs)
                # goes indented.

        self.data, self.x, self.y = data, x, y

        self.units = units
        if name is not None:
            self.name = name
        elif file is not None:
            self.name = os.path.basename(file)
        else:
            self.name = ""
        # print(name)

    def __call__(self):
        return self.data, self.x, self.y

    def __add__(self, other, *args, **kwargs):
        return Data2D(
            *sum_data(self(), other(), *args, **kwargs),
            units=self.units,
            name=self.name + " + " + other.name
        )

    def __mul__(self, scale, *args, **kwargs):
        res = self.copy()
        if len(np.shape(scale)) <= 1:  # multiply by scalar(s)
            if np.size(scale) == 3:
                res.x = scale[0] * res.x
                res.y = scale[1] * res.y
                res.data = scale[2] * res.data
            elif np.size(scale) == 1:
                res.data = res.data * scale
            else:
                raise ValueError("wrong number of elements in Data2D multiplication")
        elif len(np.shape(scale)) == 2:  # multiply by matrix
            # completely untested
            print("warning, multiplication between arrays was never tested")
            assert scale.shape == self.data.shape
            self.data = self.data * scale
        elif isinstance(scale, Data2D):
            tmp = scale.resample(self)
            res = self.copy()
            res.data = self.data * tmp.data
        else:
            raise ValueError("Multiply Data2D by wrong format!")
        return res

    def __rmul__(self, scale, *args, **kwargs):
        return self.__mul__(scale, *args, **kwargs)

    def __neg__(self):
        return self.__mul__(-1)

    def __sub__(self, other, *args, **kwargs):
        assert self.units == other.units
        res = Data2D(*subtract_data(self(), other(), *args, **kwargs), units=self.units)
        res.name = self.name + " - " + other.name
        return res

    def __truediv__(self, other):
        return self * (1.0 / other)
        
    def __repr__(self):
    
        return '<.Data2D "%s" at %s>'%(self.name,hex( id(self)))
        
        ''' NSee notes in profile_class'''

    def merge(self, other, topoints=False):
        """Return the merged data between a and b.
        If `topoints` is True points cloud data are returned.
        Gaps are brutally interpolated (unless `topoints` is set)."""

        if self.units != other.units:
            raise "incompatible units"

        p1 = self.topoints()
        p2 = other.topoints()
        res = np.vstack([p1, p2])
        if topoints:
            return res

        res = points_autoresample(res)
        return Data2D(*res, units=self.units, name=self.name + " // " + other.name)

    def plot(self, title=None, *args, **kwargs):
        """plot using data2d.plot_data and setting automatically labels and colorscales.
        by default data are filtered at 3 sigma with 2 iterations for visualization, pass nsigma = None to include all data.
        Additional arguments are passed to plot. """

        units = self.units if self.units is not None else ["","",""]
        nsigma0 = 1  # default number of stddev for color scale
        # import pdb
        # to change the default behavior
        if "stats" in kwargs:
            stats = kwargs.pop("stats")
            fmt = kwargs.pop("fmt",None)
        else: 
            stats = [[0,1,3],[6],[6]]
            # format for legend labels (replace "stdev" with "rms")
            fmt = kwargs.pop("fmt", ['mean: %.3g '+units[2],
               'rms: %.3g '+units[2],
               'PV: %.3g '+units[2],
               'size: %i X %i'])
        
        # to change the default behavior
        nsigma = kwargs.pop("nsigma", nsigma0)
        m = self.data
        #pdb.set_trace()
        res = plot_data(
            self.data,
            self.x,
            self.y,
            units=units,
            fmt=fmt,
            stats=stats,
            nsigma=nsigma,
            *args,
            **kwargs
        )
        if title is None:
            if self.name is not None:
                title = self.name
        plt.title(title)
        return res

    plot = update_docstring(plot, plot_data)

    def load(self, filename, *args, **kwargs):
        """A simple file loader using data_from_txt"""
        self.data, self.x, self.y = data_from_txt(filename, *args, **kwargs)

    load = update_docstring(load, data_from_txt)

    def save(self, filename, *args, **kwargs):
        """Save data using data2d.save_data"""
        return save_data(filename, self.data, self.x, self.y, *args, **kwargs)

    save.__doc__ = save_data.__doc__

    from functools import update_wrapper

    # @update_wrapper(rotate_data)  #this doesn't work as I would like

    def rotate(self, angle, *args, **kwargs):
        """call data2D.rotate_data, which rotate array of an arbitrary angle in degrees in direction
        (from first to second axis)."""
        res = self.copy()
        res.data, res.x, res.y = rotate_data(
            self.data, self.x, self.y, angle, *args, **kwargs
        )
        return res

    rotate = update_docstring(rotate, rotate_data)

    def rot90(self, k=1, *args, **kwargs):
        """call data2D.rotate_data, which uses numpy.rot90 to rotate array of an integer multiple of 90 degrees in direction
        (from first to second axis)."""
        res = self.copy()
        res.data, res.x, res.y = rotate_data(*res(), k=k, *args, **kwargs)

        return res

    rot90 = update_docstring(rot90, rotate_data)

    def shift(self, xoffset=None, yoffset=None, zoffset=None):
        """Shift data of given offsets along one or more axes.
        `offsets` can be provided either as 1 (data offset), 2 (x,y) or 3 separate values, or as a single 2 (x,y) or 3 elements vector."""

        # 1,2,3
        # None, 2, 3
        # None, None, 3
        # None, 2, None
        # 1, None, None
        # 1, None, 3
        # 1, 2, None
        # [1,2,3], None, None
        # [1,2,3], ... fallisce se non none
        # [1,2], None, None
        # pdb.set_trace()
        offsets = [0, 0, 0]
        if yoffset is None and zoffset is None:
            # 1, None, None
            # [1,2,3], None, None
            # [1,2], None, None
            if xoffset is not None:
                offsets[: np.size(xoffset)] = np.array(xoffset)
        else:
            assert np.size(xoffset) == 1
            offsets[0] = 0 if xoffset is None else xoffset
            offsets[1] = 0 if yoffset is None else yoffset
            offsets[2] = 0 if zoffset is None else zoffset
            # [1,2,3], ... fallisce se non none
            # [1,2], ... fallisce se non none

        # 1,2,3
        # None, 2, 3
        # None, None, 3
        # None, 2, None
        # 1, None, 3
        # 1, 2, None

        return offsets

    tv = [
        [1, 2, 3],
        [None, 2, 3],
        [None, None, 3],
        [None, 2, None],
        [1, None, None],
        [1, None, 3],
        [1, 2, None],
        [[1, 2, 3], None, None],
        [[1, 2, 3], 5, None],
        [[1, 2], None, None],
    ]

    def transpose(self):
        res = self.copy()
        res.data, res.x, res.y = transpose_data(self.data, self.x, self.y)
        return res

    transpose = update_docstring(transpose, transpose_data)

    def apply_transform(self, *args, **kwargs):
        res = self.copy()
        # pdb.set_trace()
        res.data, res.x, res.y = apply_transform_data(
            self.data, self.x, self.y, *args, **kwargs
        )
        return res

    apply_transform = update_docstring(apply_transform, apply_transform_data)

    def apply_to_data(self, func, *args, **kwargs):
        """apply a function from 2d array to 2d array to data."""
        res = self.copy()
        res.data = func(self.data, *args, **kwargs)
        return res

    def crop(self, *args, **kwargs):
        """crop data making use of function data2D.crop_data, where data,x,y are taken from a"""
        res = self.copy()
        res.data, res.x, res.y = crop_data(self.data, self.x, self.y, *args, **kwargs)
        return res

    crop = update_docstring(crop, crop_data)

    def level(self, *args, **kwargs):
        res = self.copy()
        res.data, res.x, res.y = level_data(self.data, self.x, self.y, *args, **kwargs)
        return res

    level = update_docstring(level, level_data)

    def resample(self, other, *args, **kwargs):
        """TODO, add option to pass x and y instead of other as an object."""
        res = self.copy()
        if self.units is not None and other.units is not None:
            if self.units != other.units:
                raise ValueError(
                    "If units are defined they must match in Data2D resample."
                )
        res.data, res.x, res.y = resample_data(res(), other(), *args, **kwargs)
        return res

    resample = update_docstring(resample, resample_data)

    def add_markers(self, *args, **kwargs):
        #f = plt.figure()
        self.plot()
        ax = add_clickable_markers2(*args, **kwargs)
        markers = ax.markers
        #plt.close(f)
        return markers

    def psd(
        self,
        wfun=None,
        rmsnorm=True,
        norm=1,
        analysis=False,
        subfix="",
        name=None,
        *args,
        **kwargs
    ):
        """return a PSD2D object with 2D psd of self.
        If analysis is set True, psd2d_analysis plots are generated and related parameters
          are passed as args. You need to pass also title, it generates output,
          this is subject to change, at the moment, pass empty string to generate plots
          or string to create output graphics.
        subfix and name are used to control the name of returned object.
        units are set in units of self because of the ambiguity mentioned in
        pySurf.psd2d.psd_units, and consistently with functions in `pySurf.psd2d`.
        """

        if analysis:
            title = kwargs.pop("title", (name if name is not None else ""))
            f, p = psd2d_analysis(
                self.data,
                self.x,
                self.y,
                wfun=wfun,
                norm=norm,
                rmsnorm=rmsnorm,
                title=title,
                units=self.units,
                *args,
                **kwargs
            )
        else:
            f, p = psd2d(
                self.data,
                self.x,
                self.y,
                wfun=wfun,
                norm=norm,
                rmsnorm=rmsnorm,
                *args,
                **kwargs
            )

        newname = name if name is not None else fn_add_subfix(self.name, subfix)
        return PSD2D(p, self.x, f, units=self.units, name=newname)

    psd = update_docstring(psd, psd2d)

    def remove_nan_frame(self, *args, **kwargs):
        res = self.copy()
        res.data, res.x, res.y = data2D.remove_nan_frame(
            self.data, self.x, self.y, *args, **kwargs
        )
        return res

    remove_nan_frame = update_docstring(remove_nan_frame, data2D.remove_nan_frame)

    def topoints(self):
        """convenience function to get points using matrix_to_points2."""
        return matrix_to_points2(self.data, self.x, self.y)

    def std(self, axis=None):
        """return standard deviation of data excluding nans"""
        return np.nanstd(self.data, axis=axis)

    def copy(self):
        """copy.deepcopy should work well."""
        return deepcopy(self)

    def stats(self,*args,**kwargs):
        
        units = kwargs.pop('units',self.units)
        #breakpoint()
        return get_stats(self.data,self.x,self.y,units=units,*args,**kwargs)

    def printstats(self, label=None, fmt="%3.2g"):
        if label is not None:
            print(label)
        s = ("%s PV: " + fmt + ", rms: " + fmt) % (
            self.name,
            span(self.data, size=True),
            np.nanstd(self.data),
        )
        print(s)
        return s

    def align_interactive(self, other, find_transform=find_affine):
        """interactively set markers and align self to other.
        Alignment is performed using the transformation returned by
        find_transform(markers1,markers2) after markers are interactively set.
        Return aligned Data2D object.
        There is an experimental version for dlist in scripts."""
        from pySurf.scripts.dlist import add_markers

        m1, m2 = add_markers([self, other])
        trans = find_transform(m1, m2)
        return self.apply_transform(trans)

    def remove_outliers(self, fill_value=np.nan, mask=False, *args, **kwargs):
        """use dataIO.remove_outliers to remove outliers from data. return a new Data2D object with outliers replaced by `fill_value`. If `mask` is set returns mask (easier than extracting it from returned object)."""
        res = self.copy()
        m = remove_outliers(res.data, *args, **kwargs)  # boolean mask
        # pdb.set_trace()
        if mask:
            return m
        res.data[~m] = fill_value
        return res

    remove_outliers = update_docstring(remove_outliers, outliers.remove_outliers)

    def extract_profile(self, *args, **kwargs):
        p = self.topoints()
        prof = points.extract_profile(p, *args, **kwargs)
        return prof

    extract_profile = update_docstring(extract_profile, points.extract_profile)

    def histostats(self, *args, **kwargs):
        res = data_histostats(
            self.data, self.x, self.y, units=self.units, *args, **kwargs
        )
        plt.title(self.name)
        return res

    histostats = update_docstring(histostats, data_histostats, "\n-------------\n")

    def slope(self, *args, **kwargs):
        # import pdb
        # pdb.set_trace()
        scale = kwargs.pop("scale", None)
        if self.units is not None:
            if scale is None:
                # check if x and y in mm and z in um.
                if self.units[0] == self.units[1]:
                    if self.units[0] == "mm":
                        if self.units[0] == "mm" and self.units[2] == "um":
                            scale = (1.0, 1.0, 1000.0)
                else:
                    raise ValueError("x and y different units in slope calculation")
        else:
            scale = (1.0, 1.0, 1.0)

        say, sax = slope_2D(self.data, self.x, self.y, scale=scale, *args, **kwargs)

        return Data2D(
            *sax,
            units=[self.units[0], self.units[1], "arcsec"],
            name=self.name + " xslope"
        ), Data2D(
            *say,
            units=[self.units[0], self.units[1], "arcsec"],
            name=self.name + " yslope"
        )

    slope = update_docstring(slope, slope_2D)


class PSD2D(Data2D):
    """It is a type of data 2D with customized behavoiur and additional properties
    and methods.
    """

    def __init__(self, *args, **kwargs):
        '''super is called implicitly (?non vero)
        """needs to be initialized same way as Data2D"""
        #if a surface or a wdata,x,y are passed, these are interpreted as
        super().__init__(*args,**kwargs)
        '''
        super().__init__(*args, **kwargs)

    def plot(self, *args, **kwargs):
        u = kwargs.pop("units", self.units)
        # pdb.set_trace()
        return plot_psd2d(self.y, self.data, self.x, units=u, *args, **kwargs)

    def avgpsd(self, *args, **kwargs):
        """avg, returns f and p. Can use data2D.projection keywords `span` and `expand` to return PSD ranges."""
        return self.y, projection(self.data, axis=1, *args, **kwargs)

    def rms_power(self, plot=False, rmsrange=None, *args, **kwargs):
        """Calculate rms slice power by integrating .
        If plot is set also plot the whole thing."""

        # pdb.set_trace()
        if plot:
            return plot_rms_power(
                self.y,
                self.data,
                self.x,
                units=self.units,
                rmsrange=rmsrange,
                *args,
                **kwargs
            )
        else:
            """this is obtained originally by calling rms_power, however the function deals with only scalar input for rms range.
            Part dealing with multiple ranges is in plot_rms_power, but should be moved to rms_power."""
            if rmsrange is not None:
                if np.size(rmsrange) != 2:
                    raise NotImplementedError

            return rms_power(self.y, self.data, rmsrange=rmsrange, *args, **kwargs)


def test_rot90():
    a = np.ones(250).reshape((25, 10))
    a[6:7, 6:9] = 3
    d = Data2D(a)

    plt.close("all")
    d.plot()
    plt.title("original")

    plt.figure()
    # return a,d
    c = d.rot90()
    c.plot()
    plt.title("rotated")

    plt.figure()
    # return a,d
    c = d.rot90(k=2)
    c.plot()
    plt.title("rotated k=2")

    plt.figure()
    # return a,d
    c = d.rot90(k=2, center=(10, 5))
    c.plot()
    plt.title("rotated k=2 about (10,5)")


def test_class_init(wfile=None, *args, **kwargs):
    """test init and plot"""
    from pathlib import PureWindowsPath

    from dataIO.fn_add_subfix import fn_add_subfix
    
    from pySurf.data2D import load_test_data

    # from pySurf.data2D import data, load_test

    d1, x1, y1 = load_test_data(wfile, *args, **kwargs)

    plt.figure(1)
    plt.clf()
    # plt.suptitle(relpath)
    plt.title("use plot_data function")
    plot_data(d1, x1, y1, aspect="equal")

    a = Data2D(d1, x1, y1)
    plt.figure(2)
    plt.clf()
    plt.title("From data")
    a.plot(aspect="equal")

    b = Data2D(file=wfile, ytox=220 / 1000.0, center=(0, 0))
    plt.figure(3)
    plt.clf()
    plt.title("from filename")
    b.plot(aspect="equal")

    b.save(
        os.path.join(
            outpath, os.path.basename(fn_add_subfix(relpath.as_posix(), "", ".txt"))
        ),
        makedirs=True,
    )
    b.remove_nan_frame()

    plt.figure()
    plt.title("removed nans")
    b.plot(aspect="equal")


if __name__ == "__main__":
    plt.close("all")
    test_class_init()
    test_rot90()



import numpy as np
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d as m3d

def fit_3D_line(points,plot=False):
    '''return direction vector and point for the fitting line.
    Note that it should contain also somewhere'''
    # Calculate the mean of the points, i.e. the 'center' of the cloud
    datamean = points.mean(axis=0)

    # Do an SVD on the mean-centered data.
    uu, dd, vv = np.linalg.svd(points - datamean)
    
    if (plot):
        ax=plt.gca()
        ax.scatter3D(*points.T)  #this must be     ax = m3d.Axes3D(plt.figure())
        span=((points.max(axis=0)-points.min(axis=0))/vv[0]).max()
        linepts = vv[0] * np.mgrid[-span/2:span/2:2j][:, np.newaxis]+datamean
        ax.plot3D(*linepts.T)
        plt.show()
        
    # Now vv[0] contains the first principal component, i.e. the direction
    # vector of the 'best fit' line in the least squares sense.
    return vv[0],datamean



import numpy as np
from matplotlib import pyplot as plt



def eucDist(p1,p2):
    '''demenziale distanza euclidea'''
    d=np.sqrt(np.nansum(((p1-p2)**2)))
    return d

defaultDist=eucDist

	
def distanceTable(pList,dist=defaultDist,format=None):
    '''from list of points pList, create table of distances. 
    dist is a function setting a metric, if format is format string, if provided result is converted to string using the format'''
    
    ndim=len(pList)
    t=[dist(u,v) for u in pList for v in pList]
    t=np.resize(t,(ndim,ndim))
    if format is not None:
        t=np.char.mod(format,t)
    return t


import os

def fn_add_subfix(fileName,subfix="",newext=None,strip=False,pre=""):
    """Add a subfix to a filename. A new extension can be defined (dot must be included).
    if newext is None, leaves extension unchanged (use empty string to remove extension). 
    If strip is set, only file name (not including path) is returned.
    If pre is passed, it is added as prefix (before the file name, excluding the leading directory).
    datafile='file_with_data.txt'
    outfolder='output_directory'
    print fn_add_subfix(datafile,'_output')
    print fn_add_subfix(datafile,'_output','.dat')
    import os
    fn_add_subfix(os.path.join(outfolder,datafile),'_output','.dat')"""
    
    #2014/07/28 exclude dot from extension (it must be explicitily included in the extension string).
    #   This is useful to remove extension by calling fn_add_subfix(fileName,"","") and also more
    #   consistent with os.path.splitext that returns the extension string with dot included.
    import os
    fn,ext=os.path.splitext(fileName)
    if newext!=None:
        ext=newext
    dir,base=os.path.split(fn)
    if strip:
        dir=""
    return os.path.join(dir,pre+base+subfix+ext)
    """
    if pre:
        return os.path.join(dir,subfix+base+ext)
    else:    
        return os.path.join(dir,base+subfix+ext)"""


import logging
from logging import config

def logging_function():
    """A function which prints and log info. No return value"""
    
    print("\nIn Logging Function ",logging_function)
    print("logging: ",logging,flush=True)
    logger = logging.getLogger()
    print('logger without name: %s ID: %s'%(logger,id(logger)),flush=True)
    logger.warning('Warning raised by logger')
    logger = logging.getLogger(__name__)
    print("logger %s called with name \033[1;31m%s\033[1;0m, ID: %s" %(logger,__name__,id(logger)),flush=True)
    logger.warning('Warning raised by logger')




def start_logger(logger = None,cfgfile=None):

    #if logger exists (this is based only in checking it has any handler) 
    # recover the current logger without adding handlers, if not load from file
    #  or if config file is not provided, uses default options for logger
       
    print("N: ",__name__) # qui mostra non serve perche sara start_logger    
    if logger is None:
        print("logger non defined, get one.")       
        logger=logging.getLogger()
    else:
        print("logger existing")
              
    print("logger %s, ID: %s" %(logger,id(logger)))

    if len(logger.handlers) == 0:
        """not sure about original indentation here."""   
        print("logger has no handlers.")
        if cfgfile is not None:
            print("Read logger settings from conf file ",cfgfile)
            logger=logging.config.fileConfig(cfgfile)
        else:
            print("conf file not defined, use default")
            logging.basicConfig(format='%(asctime)s %(levelname)s %(message)s',force=True)  #set default   
            print('logger: ',logger,' ID: ',id(logger))
            #print('cfgfile missing, and no logger existing, not sure what happens now.')
            
    return logger

def reset_logger(logger, full=False):
    """ close all handlersr in place. 
        If full is set, logger is also shutdown (cannot be used after this without
        being reinitialized). This strictly needs to be done only after program execution."""
        
    for h in logger.handlers:
        h.close()
    if full: 
        logger.handlers.clear()
        logger.shutdown()

# legacy        
class Logger():
    
    #methods that mimic the behavior of logging.Logger
    def info(self,msg):
        print (msg)
    def debug(self,msg):
        print(msg)
        


cfgfile='logging.ini'

'''
# from http://techies-world.com/how-to-redirect-stdout-and-stderr-to-a-logger-in-python/
# To redirect stdout and stderr to a logger, use the following snippet.
# 

class StreamToLogger(object):
      """
      Fake file-like stream object that redirects writes to a logger instance.
      """
      def __init__(self, logger, log_level=logging.INFO):
            self.logger = logger
            self.log_level = log_level
            self.linebuf = ''

      def write(self, buf):
            for line in buf.rstrip().splitlines():
                  self.logger.log(self.log_level, line.rstrip())

logging.basicConfig(
level=logging.DEBUG,
format='%(asctime)s:%(levelname)s:%(name)s:%(message)s',
filename="out.log",
filemode='a'
)

stdout_logger = logging.getLogger('STDOUT')
sl = StreamToLogger(stdout_logger, logging.INFO)
sys.stdout = sl

stderr_logger = logging.getLogger('STDERR')
sl = StreamToLogger(stderr_logger, logging.ERROR)
sys.stderr = sl

print "Test to standard out"
raise Exception('Test to standard error')

'''





from matplotlib import pyplot as plt
import numpy as np
from cycler import cycler
from collections import OrderedDict

from pyProfile.psd import make_psd_plots as make_plots
print ("WARNING: makenicePSDplots was moved in pyProfile.psd, it will be discontinued, modify your import.")

"""
def make_plots(toplot,outfile=None): 
    plt.figure()
    # 1. Setting prop cycle on default rc parameter
    plt.rc('lines', linewidth=1.5)
    plt.rc('axes', prop_cycle=(cycler('linestyle', ['-', '--', ':', '-.'])* cycler('color', ['r', 'g', 'b', 'y']) ))
       
    for ff,ll in toplot.items():    
        f,p=np.genfromtxt(ff,delimiter='',unpack=True,usecols=[0,1])
        plt.plot(f,p,label=ll)
        plt.loglog()
        plt.xlabel('freq. (mm$^-1$)')
        plt.ylabel('mm $\mu$m$^2$')
        plt.grid(1)
        plt.legend(loc=0)
        plt.show()
        
    if outfile is not None:
        plt.savefig(outfile)
        print("results saved in %s"%outfile)
        #plt.savefig(fn_add_subfix(ff,'_psdcomp','.png'))
"""


"""
Plot_surface_analysis and leveling.

functions to plot and compare data, to be removed.
"""


import numpy as np
from matplotlib import pyplot as plt
from pySurf.readers.instrumentReader import matrixZygo_reader
from pySurf.data2D import plot_data,removelegendre,level_by_line,projection
from pySurf.data2D import levellegendre,save_data,get_data
# from pySurf.psd2d import psd2d_analysis,rms_power
from utilities.imaging.fitting import legendre2d
from dataIO.fn_add_subfix import fn_add_subfix
from dataIO.span import filtered_span
from dataIO.outliers import remove_outliers
from plotting.multiplots import diff_images
from plotting.fignumber import fignumber
from dataIO.span import span
import os
import pdb
from scipy.stats import ttest_ind
# from pySurf.data2D import plot_surface_analysis
# from pySurf.psd2d import psd_variability
from  plotting.backends import maximize
from plotting.multiplots import compare_images

def plot_surface_analysis(wdata,x,y,label="",outfolder=None,nsigma_crange=1,fignum=None,
    figsize=(9.6,6.7),psdrange=None,levelingfunc=None,frange=None):
    """Create two panel preview plots for data. 
    
    This is two panels of raw and filtered data (second order removed) + histogram 
       of height distribution + PSD. 
    Label is typically the filename and is used to generate title and output filenames.
    fignum and figsize are for PSD2D figure (fignum=0 clear and reuse current figure, 
    None create new).
    nsigma_crange and levelingfunc are used to evaluate color range for plot by iteratively 
    leveling and removing outliers at nsigma until a convergence is reached.
    psdrange is used 
    frange is used for plot of PSD2D
    """
    from pySurf.psd2d import psd2d_analysis
    
    units=['mm','mm','$\mu$m']
    order_remove=2 #remove sag, this is the order that is removed in leveled panel
    if np.size(units)==1:   #unneded, put here just in case I decide to take units
        units=np.repeat(units,3)  #as argument. However ideally should be put in plot psd function.
    
    if levelingfunc is None:
        levelingfunc=lambda x: x-legendre2d(x,1,1)[0]
    
    if outfolder is not None:
        os.makedirs(os.path.join(outfolder,'PSD2D'),exist_ok=True)
    else:
        print('OUTFOLDER is none')
    
    wdata=levelingfunc(wdata)  #added 20181101
    
    rms=np.nanstd(wdata)
    #crange=remove_outliers(wdata,nsigma=nsigma_crange,itmax=1,range=True)
    crange=remove_outliers(wdata,nsigma=nsigma_crange,
        flattening_func=levelingfunc,itmax=2,span=1)
    
    plt.clf()
    
    #FULL DATA
    plt.subplot(221)
    #plt.subplot2grid((2,2),(0,0),1,1)
    plot_data(wdata,x,y,units=units,title='leveled data',stats=True)
    plt.clim(*crange)
    
    #SUBTRACT LEGENDRE
    ldata=levellegendre(y,wdata,order_remove)
    plt.subplot(223)
    lrange=remove_outliers(ldata,nsigma=nsigma_crange,itmax=1,span=True)
    #plt.subplot2grid((2,2),(1,0),1,1)
    plot_data(ldata,x,y,units=units,title='%i orders legendre removed'%order_remove,stats=True)
    plt.clim(*lrange) 
    
    #HISTOGRAM
    #plt.subplot2grid((2,2),(0,1),2,1)
    plt.subplot(222)
    plt.hist(wdata.flatten()[np.isfinite(wdata.flatten())],bins=100,label='full data')
    plt.hist(ldata.flatten()[np.isfinite(ldata.flatten())],bins=100,color='r',alpha=0.2,label='%i orders filtered'%order_remove)
    #plt.vlines(crange,*plt.ylim())
    plt.vlines(crange,*plt.ylim(),label='plot range raw',linestyle='-.')
    plt.vlines(lrange,*plt.ylim(),label='plot range lev',linestyle=':')
    plt.legend(loc=0)
    
    #FREQUENCY AND PSD ANALYSIS
    if fignum==0:   #this is the figure created by psd2d_analysis
        plt.clf()
        fig = plt.gcf()
    #pdb.set_trace()
    stax=plt.gca()   #preserve current axis
    f,p=psd2d_analysis(ldata,x,y,title=label,
        wfun=np.hanning,fignum=fignum,vrange=lrange,
        prange=psdrange,units=units,
        rmsrange=frange)#,ax2f=[0,1,0,0])
    plt.ylabel('slice rms ($\mu$m)')
    ps=projection(p,axis=1,span=True) #avgpsd2d(p,span=1)      
    if outfolder:
        maximize()
        plt.savefig(os.path.join(outfolder,'PSD2D',
            fn_add_subfix(label,"_psd2d",'.png')))
        save_data(os.path.join(outfolder,'PSD2D',fn_add_subfix(label,"_psd2d",'.txt')),
            p,x,f)
        np.savetxt(os.path.join(outfolder,fn_add_subfix(label,"_psd",'.txt')),np.vstack([f,ps[0],ps[1],ps[2]]).T,
            header='f(%s^-1)\tPSD(%s^2%s)AVG\tmin\tmax'%(units[1],units[2],units[1]),fmt='%f')
            #pdb.set_trace()
    plt.sca(stax)
    
    plt.subplot(224)
    plt.ylabel('axial PSD ('+units[2]+'$^2$ '+units[1]+')')
    plt.xlabel('Freq. ('+units[1]+'$^{-1}$)')
    plt.plot(f,ps[0],label='AVG')
    plt.plot(f,ps[1],label='point-wise min')
    plt.plot(f,ps[2],label='point-wise max')
    plt.plot(f,p[:,p.shape[1]//2],label='central profile')
    plt.ylim(psdrange)
    plt.loglog()
    plt.grid(1)
    plt.legend(loc=0) 
    #ideally to be replaced by:
    """
    plt.subplot(224)
    plot_psd(((f,ps[0]),(f,ps[1]),(f,ps[2),(f,p[:,p.shape[1]//2])),
            ['AVG','point-wise min','point-wise max','central profile'])
    plt.ylabel('axial '+plt.ylabel)
    plt.ylim(psdrange)
    """

    plt.suptitle('%s: rms=%5.3f 10$^{-3}$'%(label,rms*1000)+units[2])
    plt.tight_layout(rect=[0, 0.03, 1, 0.95])
    
    if outfolder:
        plt.savefig(os.path.join(outfolder,fn_add_subfix(label,"",'.png')))
    
    return (ldata,x,y),(f,p)
    
def leveldata(wdata,xwg,ywg): 
    """plot and return matrices of wdata, lwdata, lwdata2
    from data wdata. use polynomial fit, non legendre, each line is leveled on 
    extremes, not on zero piston/tilt""" 
    #import pdb
    #pdb.set_trace()
    from pySurf.data2D import removesag
    
    #create data
    lwdata=level_by_line(wdata.copy())
    lwdata2=level_by_line(wdata.copy(),removesag)
    lwdata2=level_by_line(lwdata2)  ##why this? to level extremes, no piston. The default action of level_by_line is remove line through extremes.
    
    #make plots
    #generator of axis
    for ax in compare_images([wdata,lwdata,lwdata2],xwg,ywg,titles=['original','line lev.','sag. lev.'],fignum=0)  :
        ax.set_xlabel('X (mm)')
        ax.set_ylabel('Y (mm)')

    return wdata,lwdata,lwdata2


from utilities.imaging.userKov.pySurf.points import *
#from userKov.calibrate_align import *
import numpy as np
from matplotlib import pyplot as plt
from scipy.optimize import minimize

def closest_point_on_line(points,lVersor,lPoint):
    """From a list of points in 3D space as Nx3 array, returns a Nx3 array with the corresponding closest points on the line."""
    vd=lVersor/((np.array(lVersor)**2).sum())  #normalize vector, not sure it is necessary.
    return lPoint+(points-lPoint).dot(vd)[:,np.newaxis]*(vd)

def cylinder_error3(odr=(0,0,0,0),points=None,extra=False,xy=False):  
#from v1
#origin=(0,0,0),direction=(0,1.,0),radius is calculated from best fit
    """Given a set of N points in format Nx3, returns the rms surface error on the cylinder defined by origin (intercept of the axis with x=0) and direction, 
    passed as 4-vector odr (origin_y,origin_z,direction_x,direction_z). 
    Best fit radius for odr is calculated as average. 
    If extra is set, additional values are returned :
        radius: best fit radius for the cylinder.
        deltaR[N,3]: deviation from radius for each point.
    """
    #ca 400 ms per loop
    if xy:
        origin=(0,odr[0],odr[1])
        direction=(1.,odr[2],odr[3])
    else:
        origin=(odr[0],0,odr[1])
        direction=(odr[2],1.,odr[3])
    
    #vd=direction/((np.array(direction)**2).sum())  #normalize vector, not sure it is necessary.
    x,y,z=np.hsplit(points,3)
    Paxis=closest_point_on_line(points,direction,origin)
    #origin+(points-origin).dot(vd)[:,np.newaxis]*(vd) #points on the cylinder axis closer to points to fit
    R=np.sqrt(((points-Paxis)**2).sum(axis=1)) #distance of each point from axis
    radius=R.mean()  #
    #plt.plot(R-radius)
    deltaR=(R-radius)  #square difference from expected radius for each point
    fom=np.sqrt(((deltaR)**2).sum()/len(deltaR))
    residuals=np.hstack([x,y,deltaR[:,None]])
    if extra: return fom,residuals,radius
    else: return fom  #,deltaR,radius


if __name__=="__main__":
    #always run this
    gfile='/home/rallured/Dropbox/WFS/SystemAlignment/Bonding/150529BottomLeft_Center.txt'
    fit_func=cylinder_error3

    pts=get_points(gfile,matrix=1, xrange=[0,127],yrange=[0,127])
    pts=crop_points(pts,[23,80],[15,93])
    pts[:,2]=pts[:,2]/1000.   #data are um, convert to mm before fit

    #level plane
    pts,pl=level_points(pts,returnPars=1)
    pts[:,2]=pts[:,2]-pts[np.argsort(pts[:,2])[0],2]

    pts[:,2]=-pts[:,2] #invert concavity

    plt.figure(1)
    plt.clf()
    plot_points(pts,shape=[50,50])

    def p(x): print(x) # callback function 
        #passed to minimization to print each step.

    odr2=(0,5000.,0,0.) #use nominal value for guess direction
    result=minimize(fit_func,x0=(odr2),
                    args=(pts,),options={'maxiter':500},method='Nelder-Mead',callback=p)
    print('-----------------------------------')
    print('Results of fit on region:')
    print(result)    
    odr=result.x
    fom,deltaR,coeff=fit_func(odr,pts,extra=True)
    print('Angle of cylinder axis with y axis (deg):')
    print(np.arccos(np.sqrt(1-(odr[-2:]**2).sum()))*180/np.pi)
    print('Cylinder parameters (angle(deg), R@y=0):')
    print(np.arctan(coeff)*180/np.pi,coeff)

    plt.figure(2)
    plt.clf()
    plot_points(deltaR,shape=[50,50])
    plt.show()

    



#!/usr/bin/env python
# coding: utf-8

# In[1]:


#2019/04/01 rerun this analysis that was made in days after measurement 19/03/07
#  and  not completed because of issues with python.
# C1S04 was measured in three occasions, this is the first one:
# 2019/03/07  07_ steps 25x1000, 400 Hz  might have moved on support
# 2019/07/11  14_ steps 25x1000, 400 Hz looks good
# 2019/07/11  19_ steps 25x250, 400 Hz  higher sampling

#---
#---
#2018/08/18 C1S05 after compensating layer IrC 
#   this first set of data have some displacement in corners,
#   a second set of data is reacquired on 09/19 with a better correspondance.

#2018/08/07 from C1S05_PZT_0523_analysis

# This is remake with PZT sample after it was sent back to SAO because all previous coating studies
#  were made on the uncoated sample. Since then it was coated with gold and measured today.
# 
from __future__ import print_function
get_ipython().run_line_magic('reset', '')
get_ipython().run_line_magic('load_ext', 'autoreload')
get_ipython().run_line_magic('autoreload', '2')
get_ipython().run_line_magic('pylab', '')


# In[2]:


from pySurf.points import plot_points,matrix_to_points2
from pySurf.points import subtract_points,level_points
from pySurf.points import points_find_hull,crop_points,save_points,points_autoresample
from pySurf.psd2d import calculatePSD #,plotPSDs,PSDplot
from pyProfile.profile import polyfit_profile, level_profile
from plotting.multiplots import compare_images
from dataIO.fn_add_subfix import fn_add_subfix
from dataIO.span import span
import os
from utilities.imaging import fitting as fit
from plotting.add_clickable_markers import add_clickable_markers




#gp=get_points(df,scale=gscale,center=(0,0))


pts1=get_points(fn_add_subfix(datafile,'_shape','.dat'),delimiter=' ')


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5

pts.shape


# In[174]:


a=np.array(np.arange(12).reshape((3,4)))
print('a, shape ',a.shape,':\n',a)

aa=a+np.array([float('.%02i'%i) for i in np.arange(12)]).reshape(a.shape)
print('aa, shape ',aa.shape,':\n',aa)

#this concatenates arrays on an arbitrary axis, defaults to 0
b=np.stack([aa,aa+100])
print('b, shape ',b.shape,':\n',b)


# In[ ]:


#indexing address on first axis
b[0].shape
Out[173]: (3, 4)


# In[172]:


#in numpy min and max, axis that is passed as argument is removed.
b.shape
Out[171]: (2, 3, 4)

np.min(b,axis=0).shape
Out[168]: (3, 4)

np.min(b,axis=1).shape
Out[169]: (2, 4)

np.min(b,axis=2).shape
Out[170]: (2, 3)


# In[187]:


span(b,axis=0).shape


# In[191]:


span(b,axis=0).shape
Out[188]: (3, 4, 2)

span(b,axis=1).shape
Out[189]: (2, 4, 2)

span(b,axis=2).shape
Out[190]: (2, 3, 2)


# In[210]:


span(b,axis=2)


# In[196]:


span(aa,axis=0).shape
Out[194]: (4, 2)

span(aa,axis=1).shape
Out[195]: (3, 2)


# In[205]:


s=span(pts,axis=0)
print('s, shape ',s.shape,':\n',s)
print('s[0], shape ',s[0].shape,':\n',s[0])


# In[ ]:




#This submodule includes routines to manipulate image arrays
import numpy as np
from scipy.interpolate import griddata
from utilities.imaging.fitting import legendre2d

import pdb

def unpackimage(data,xlim=[-1,1],ylim=[-1,1],remove=True):
    """Convert a 2D image into x,y,z coordinates.
    x will relate to 2nd index in order to correspond to abscissa in imshow
    y will relate to 1st index in order to correspond to oordinate in imshow
    if remove is True, NaNs will not be returned in the list of coordinates
    """
    x,y = np.meshgrid(np.linspace(xlim[0],xlim[1],np.shape(data)[1]),\
                   np.linspace(ylim[0],ylim[1],np.shape(data)[0]))
##    y = y.flatten()
##    x = x.flatten()
    
    if remove is True:
        ind = np.invert(np.isnan(data.flatten()))
        return x.flatten()[ind],y.flatten()[ind],data.flatten()[ind]

    return x.flatten(),y.flatten(),data.flatten()

def autoGrid(d,xr=[-1,1],yr=[-1,1]):
    """
    Create a meshgrid based on the shape of the image d
    """
    xspan = np.linspace(xr[0],xr[1],np.shape(d)[1])
    yspan = np.linspace(yr[0],yr[1],np.shape(d)[0])
    return np.meshgrid(xspan,yspan)
    

def shiftNaN(img,n=1,axis=0):
    """This function shifts an image in a NaN padded array
    Specify which axis to shift, and specify wh
    ich direction
    """
    #Construct array to insert
    if axis is 0:
        ins = np.repeat(np.nan,np.abs(n)*\
                     np.shape(img)[1]).reshape(np.abs(n),np.shape(img)[1])
    else:
        ins = np.repeat(np.nan,np.abs(n)*\
                     np.shape(img)[0]).reshape(np.abs(n),np.shape(img)[0])
    #If direction=0, shift to positive
    if n > 0:
        img = np.delete(img,np.arange(np.shape(img)[1]-\
                                      n,np.shape(img)[1]),axis=axis)
        img = np.insert(img,0,ins,axis=axis)
    else:
        n = np.abs(n)
        img = np.delete(img,np.arange(n),axis=axis)
        img = np.insert(img,-1,ins,axis=axis)
    return img

def padNaN(img,n=1,axis=0):
    """Pads an image with rows or columns of NaNs
    If n is positive, they are appended to the end of
    the specified axis. If n is negative, they are
    appended to the beginning
    """
    #Construct array to insert
    if axis is 0:
        ins = np.repeat(np.nan,np.abs(n)*\
                        np.shape(img)[1]).reshape(np.abs(n),np.shape(img)[1])
    else:
        ins = np.repeat(np.nan,np.abs(n)*\
                        np.shape(img)[0]).reshape(np.abs(n),np.shape(img)[0])
        ins = np.transpose(ins)
    #If direction=0, shift to positive
    if n < 0:
        img = np.concatenate((ins,img),axis=axis)
    else:
        img = np.concatenate((img,ins),axis=axis)
    return img

def padRect(img,nan_num = 1):
    """Pads an image with an outer NaN rectangle of width nan_num"""
    img = padNaN(img,n=nan_num,axis=0)
    img = padNaN(img,n=-nan_num,axis=0)
    img = padNaN(img,n=nan_num,axis=1)
    img = padNaN(img,n=-nan_num,axis=1)
    return img

def borderFill(img,n = 1,fill_value = np.NaN):
    img[:n],img[-n:] = fill_value,fill_value
    img[:,:n],img[:,-n:] = fill_value,fill_value
    return img

def tipTiltPiston(img,piston,tip,tilt,tx=None,ty=None):
    """This function adds a constant and
    tip and tilt to an array
    This makes use of tilt arrays tx,ty
    If not provided, compute using meshgrid
    Updated
    """
    if tx is None:
        ty,tx = np.meshgrid(np.arange(np.shape(img)[1]),\
                                np.arange(np.shape(img)[0]))
        tx = (tx-np.mean(tx)) / tx.max()
        ty = (ty-np.mean(ty)) / ty.max()

    return img + piston + tip*tx + tilt*ty

def nearestNaN(arr,method='nearest'):
    """Fill the NaNs in a 2D image array with the griddata
    nearest neighbor interpolation"""
    ishape = np.shape(arr)
    #Unpack image both with and without removing NaNs
    x0,y0,z0 = unpackimage(arr,remove=False)
    x1,y1,z1 = unpackimage(arr,remove=True)

    #Interpolate onto x0,y0 grid
    newarr = griddata((x1,y1),z1,(x0,y0),method=method)

    return newarr.reshape(ishape)

def rebin(a,shape):
    sh = shape[0],a.shape[0]//shape[0],shape[1],a.shape[1]//shape[1]
    return nanmean(nanmean(a.reshape(sh),axis=3),axis=1)

def stripnans(d1,removeAll=False):
    """
    Need to fix removeAll. Likely need to remove rows/columns
    in a circular fashion until all perimeter NaNs are eliminated.
    """
    d = np.copy(d1)
    if len(np.shape(d)) is 1:
        return d[~np.isnan(d)]
    if removeAll is False:
        newsize = np.shape(d)[1]
    else:
        newsize = 1
    while sum(np.isnan(d[0]))>=newsize:
        d = d[1:]
    while sum(np.isnan(d[-1]))>=newsize:
        d = d[:-1]
    if removeAll is False:
        newsize = np.shape(d)[0]
    else:
        newsize = 1
    while sum(np.isnan(d[:,0]))>=newsize:
        d = d[:,1:]
    while sum(np.isnan(d[:,-1]))>=newsize:
        d = d[:,:-1]
    return d

def transformation(x,y,r=0.,tx=0.,ty=0.):
    """Return x and y vectors after applying a rotation about
    the origin and then translations in x and y
    """
    x,y = np.cos(r)*x+np.sin(r)*y,-np.sin(r)*x+np.cos(r)*y
    x,y = x+tx,y+ty
    return x,y

def rotateImage(img,rot):
    """Apply a rotation about the center of an image using
    griddata
    """
    sh = np.shape(img)
    x,y = np.meshgrid(np.linspace(-1,1,sh[1]),np.linspace(-1,1,sh[0]))
    dx = 2./(sh[1]-1)
    dy = 2./(sh[0]-1)
    x,y = transformation(x,y,r=rot)
    x2,y2 = np.meshgrid(np.arange(x.min(),x.max()+dx,dx),\
                        np.arange(y.min(),y.max()+dy,dy))
    #Interpolate from x,y to x2,y2
    img2 = griddata((x.flatten(),y.flatten()),img.flatten(),(x2,y2))
    return stripnans(img2)

def newGridSize(img,newshape,method='linear'):
    """
    Interpolate an image onto a new shape size
    """
    shape1 = np.shape(img)
    x1,y1 = np.meshgrid(list(range(shape1[1])),list(range(shape1[0])))
    x2,y2 = np.meshgrid(np.linspace(0,shape1[1]-1,newshape[1]),\
                        np.linspace(0,shape1[0]-1,newshape[0]))
    img2 = griddata((x1.flatten(),y1.flatten()),img.flatten(),(x2,y2),\
                    method=method)
    return img2

def nanflatten(img):
    """
    Automatically remove NaNs when flattening an image
    """
    d = img.flatten()
    d = d[~np.isnan(d)]
    return d

def removePoly(l,order=2):
    """
    Remove a polynomial up to order from a slice.
    NaNs are ignored.
    """
    if np.sum(~np.isnan(l))<2:
        return
    ind = np.invert(np.isnan(l))
    x = np.arange(len(l))
    fit = np.polyfit(x[ind],l[ind],order)

    return l - np.polyval(fit,x)

def removeLegSlice(din,order=2,axis=0):
    """
    Remove a Legendre polynomial up to order from each
    slice of an image.
    """
    d = np.copy(din)
    if axis is 0:
        d = np.transpose(d)
    for i in range(len(d)):
        d[i,:] = removePoly(d[i,:],order=order)
    if axis is 0:
        d = np.transpose(d)
    
    return d

def remove2DLeg(din,xo=2,yo=0):
    """
    Remove a 2D Legendre fit to din up to
    xo and yo.
    """
    f = legendre2d(din,xo=xo,yo=yo)[0]
    
    return din-f

def removeDS9Regions(img,filename):
    """
    Read in an SAOImage region file and set all
    pixels within regions to NaN.
    File should look like:
    circle(x,y,rad)
    box(x,y,dx,dy,0)
    ellipse(x,y,dx,dy,0)
    """
    #Construct index grid
    y,x = np.meshgrid(list(range(np.shape(img)[1])),list(range(np.shape(img)[0])))
    
    #Get region data
    f = open(filename,'r')
    lines = f.readlines()

    for l in lines:
        t = l.split('(')[0]
        n = np.array(l.split('(')[1].split(','))
        # Note: this can throw an error based on Mac vs. Windows -- if there's a carriage return
        # plus a new line, the second value should be -3.
        n[-1] = n[-1][:-2]
        n = n.astype('float')
        if t == 'circle':
            ind = (x-n[1])**2+(y-n[0])**2 < n[2]**2
        elif t == 'box':
            yind = np.logical_and(y<n[0]+n[2]/2,y>n[0]-n[2]/2)
            xind = np.logical_and(x<n[1]+n[3]/2,x>n[1]-n[3]/2)
            ind = np.logical_and(xind,yind)
        elif t == 'ellipse':
            ind = []
        img[ind] = np.nan

        
    return img



print ("module moved, modify your import to use readers.instrumentReader")


# from .readers import instrumentReader 
from pySurf.readers.instrumentReader import *

'''
import sys
#del sys.modules['instrumentReader']
#sys.modules['instrumentReader'] = __import__('Mod_2')

sys.modules['instrumentReader'] = __import__('pySurf.readers.instrumentReader')
#return ir
'''

'''
def span(*args,**kwargs):
    from dataIO.span import span
    print "module span was moved to dataIO, please update inport in code as from dataIO.span import span."
    return span (*args,**kwargs)
'''


#!/usr/bin/env python

'''
Fast Plotter for Large Images - Resamples Images to a target resolution on each zoom.
Example::
    sz = (10000,20000) # rows, cols
    buf = np.arange(sz[0]*sz[1]).reshape(sz)
    extent = (100,150,1000,2000)
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    im = FastImshow(buf,extent,ax)
    im.show()
    plt.show()
'''

import numpy as np
import matplotlib.pyplot as plt


class FastImshow:
    '''
    Fast plotter for large image buffers
    Example::
        sz = (10000,20000) # rows, cols
        buf = np.arange(sz[0]*sz[1]).reshape(sz)
        extent = (100,150,1000,2000)
        fig = plt.figure()
        ax  = fig.add_subplot(111)
        im = FastImshow(buf,extent,ax)
        im.show()
        plt.show()
    '''
    def __init__(self,buf,ax,extent=None,tgt_res=512):
        '''
        [in] img buffer
        [in] extent
        [in] axis to plot on
        [in] tgt_res(default=512) : target resolution
        '''
        self.buf = buf
        self.sz = self.buf.shape
        self.tgt_res = tgt_res
        self.ax = ax

        # Members required to account for mapping extent to buf coordinates
        if extent:
            self.extent = extent
        else:
            self.extent = [ 0, self.sz[1], 0, self.sz[0] ]
        self.startx = self.extent[0]
        self.starty = self.extent[2]
        self.dx = self.sz[1] / (self.extent[1] - self.startx ) # extent dx
        self.dy = self.sz[0] / (self.extent[3] - self.starty ) # extent dy

    # end __init__

    def get_strides( self,xstart=0, xend=-1, ystart=0, yend=-1, tgt_res=512 ):
        '''
        Get sampling strides for a given bounding region. If none is provided,
           use the full buffer size
        '''
        # size = (rows,columns)
        if xend == -1:
            xend = self.sz[1]
        if yend == -1:
            yend = self.sz[0]
        if (xend-xstart) <= self.tgt_res:
            stridex = 1
        else:
            stridex = max(int((xend - xstart) / self.tgt_res),1)

        if (yend-ystart) <= self.tgt_res:
            stridey = 1
        else:
            stridey = max(int((yend - ystart) / self.tgt_res),1)

        return stridex,stridey
    # end get_strides

    def ax_update(self, ax):
        '''
        Event handler for re-plotting on zoom
        - gets bounds in img extent coordinates
        - converts to buffer coordinates
        - calculates appropriate strides
        - sets new data in the axis
        '''
        ax.set_autoscale_on(False)  # Otherwise, infinite loop

        # Get the range for the new area
        xstart, ystart, xdelta, ydelta = ax.viewLim.bounds
        xend = xstart + xdelta
        yend = ystart + ydelta

        xbin_start = int(self.dx * ( xstart - self.startx ))
        xbin_end   = int(self.dx * ( xend - self.startx ))
        ybin_start = int(self.dy * ( ystart - self.starty ))
        ybin_end   = int(self.dy * ( yend - self.starty ))

        # Update the image object with our new data and extent
        im = ax.images[-1]

        stridex,stridey = self.get_strides( xbin_start,xbin_end,ybin_start,ybin_end)

        im.set_data( self.buf[ybin_start:ybin_end:stridey,xbin_start:xbin_end:stridex] )

        im.set_extent((xstart, xend, ystart, yend))

        ax.figure.canvas.draw_idle()
    # end ax_update

    def show(self):
        '''
        Initial plotter for buffer
        '''
        stridex, stridey = self.get_strides()
        self.ax.imshow( buf[::stridex,::stridey],extent=self.extent,origin='lower',aspect='auto' )
        self.ax.figure.canvas.draw_idle() 

        self.ax.callbacks.connect('xlim_changed', self.ax_update)
        self.ax.callbacks.connect('ylim_changed', self.ax_update)
    # end show

# end ImgDisplay

if __name__=="__main__":
    sz = (10000,20000) # rows, cols
    buf = np.arange(sz[0]*sz[1]).reshape(sz)
    extent = (100,150,1000,2000)
    fig = plt.figure()
    ax  = fig.add_subplot(111)
    im = FastImshow(buf,ax,extent=extent,tgt_res=1024)
    im.show()

    plt.show()
	
	
	
	



import sys
import numpy as np
import sys
from matplotlib import pyplot as plt

import matplotlib.patches as patches

#2017/08/03 made some adjustments, mostly in interactivity. However plt.ion/ioff don't update immediately,
#  it is necessary to close the window.

def add_mask(fig=None,ax=None,key='enter',propertyname='mask',modkey='control',hold=False):
    """Select a rectangle on image."""
    
    #useful references  
    #http://astronomy.nmsu.edu/holtz/a575/ay575notes/node6.html
    #https://matplotlib.org/examples/event_handling/keypress_demo.html
    drawing=False
    
    def press(event):
        #print('press', event.key)
        sys.stdout.flush()
        if event.key == key:
            fig.canvas.mpl_disconnect(cid1)
            fig.canvas.mpl_disconnect(cid2)
            #fig.canvas.stop_event_loop()
            print (plt.isinteractive())
            plt.interactive(interact_store)  #this sets interactive to original status, but the behaviour doesn't update (if set to ON you still need to close the figure)
            #plt.show(block=False)
            print (plt.isinteractive())
        
    def onclick(event):
        button=event.button
        x=event.xdata
        y=event.ydata
        fig=event.canvas.figure
        ax=event.inaxes

        if button==1: 
            if event.key is not None:
                if event.key == modkey:

                        
                    l=getattr(fig,propertyname)
                    l.append([event.xdata, event.ydata])
                    #add point
                    ax.set_autoscale_on(False)
                    ax.plot(x,y,'or',picker=5,markerfacecolor=None)
        if button==3:
            if event.key is not None:
                if event.key == modkey:
                    l=getattr(fig,propertyname)
                    if len(l)>0:
                        dist=((np.array(l)-(x,y))**2).sum(axis=1)
                        pop=l.pop(np.argmin(dist))
                        dist2=[(np.array((pop[0]-a.get_xdata(),pop[1]-a.get_ydata()))**2).sum()  for a in ax.lines]
                        ax.lines.pop(np.argmin(dist2))
        
        plt.draw()
        #print 'button=%d, x=%d, y=%d, xdata=%f, ydata=%f'%(
        #event.button, event.x, event.y, event.xdata, event.ydata)
    
    def handle_close(evt):
        print('Closed Figure!')
        fig.canvas.mpl_disconnect(cid3)

    #to disconnect:
    #fig.canvas.mpl_disconnect(cid)
    """
    def on_pick(event):
        # not used. this is an alternative way, it automatically manage the selection of the plotted point.
        import matplotlib
        thisline = event.artist
        if isinstance(thisline,matplotlib.lines.Line2D):
            #code to remove point from markers list
            thisline.remove()
            plt.show()
    """
    
    if fig is None:
        fig=plt.gcf()
    interact_store=plt.isinteractive()
    #plt.show(block=True) #This makes impossible to select the points because it blocks events handling.
    #plt.show(block=False) #This works to select the points, but it doesn't release afterwards.
    if hold:
        print ('hold')
        plt.ioff()
    else:
        plt.ion()  #this works in setting interactive mode off independently on starting status. However it sets the status forever, setting ioff here and ion at the end alters
    # the status variable, but it is not applied to the window.
    # so it must be set off here if you want to proceed at the end, but it is necessary to 
    # close the windows to continue.
    
    #fig.canvas.start_event_loop(20)  #this is useful only for timed events (ginput?) documentation is
        #not  very clear to this regar, but it's clearly useless here.
    #add a list property (empty) with the chosen name for markers,
    #if property already exist it will automatically append.    
    if not hasattr(fig,propertyname):
        setattr(fig,propertyname,[])
        
    #ax.imshow(im)
    #ax.autoscale(False)
    cid1 = fig.canvas.mpl_connect('button_press_event', onclick)
    cid2 = fig.canvas.mpl_connect('key_press_event', press)
    cid3 = fig.canvas.mpl_connect('close_event', handle_close)
    return fig
    

      
if __name__=="__main__":


    from matplotlib.widgets import RectangleSelector
    #from generalPlotting.pickPointList import add_clickable_markers
    imfile=r'test\input_data\IMG_4509_crop.jpg'
    a=plt.imread(imfile)
    plt.clf()
    plt.imshow(a)
    
    def line_select_callback(eclick, erelease):
        'eclick and erelease are the press and release events'
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata
        print("(%3.2f, %3.2f) --> (%3.2f, %3.2f)" % (x1, y1, x2, y2))
        print(" The button you used were: %s %s" % (eclick.button, erelease.button))


    def toggle_selector(event):
        print(' Key pressed.')
        if event.key in ['Q', 'q'] and toggle_selector.RS.active:
            print(' RectangleSelector deactivated.')
            toggle_selector.RS.set_active(False)
        if event.key in ['A', 'a'] and not toggle_selector.RS.active:
            print(' RectangleSelector activated.')
            toggle_selector.RS.set_active(True)
        # drawtype is 'box' or 'line' or 'none'
        
    toggle_selector.RS = RectangleSelector(plt.gca(), line_select_callback,
                                           drawtype='box', useblit=True,
                                           button=[1, 3],  # don't use middle button
                                           minspanx=5, minspany=5,
                                           spancoords='pixels',
                                           interactive=True)
    plt.connect('key_press_event', toggle_selector)
    plt.show()


"""Artificially generated test surfaces.
Can be incorporated in readers."""

import numpy as np
import matplotlib.pyplot as plt
from dataIO.span import span

def makeGaussian(size, fwhm = 3, center=None):
    """ Make a square gaussian kernel.
    size is the length of a side of the square
    fwhm is full-width-half-maximum, which
    can be thought of as an effective radius.
    """

    x = np.arange(0, size, 1, float)
    y = x[:,np.newaxis]
    
    if center is None:
        x0 = y0 = size // 2
    else:
        x0 = center[0]
        y0 = center[1]
    
    return np.exp(-4*np.log(2) * ((x-x0)**2 + (y-y0)**2) / fwhm**2)



def make_sag(nx,ny):
    """ create surface with sag along y with peak-to-valley 1.
    Note that this is the range of data, not of the surface (i.e. if 
    number of points in y is even, the analytical minimum of the surface lies
    between the two central pixels and it is not included in data)"""
    
    ss=(np.arange(ny)-(ny-1)/2.)**2 #makes a parabola centered on 0
    ss=ss/span(ss,size=1)-np.min(ss) #rescale height to 1
    sag=np.repeat(ss[np.newaxis, :], nx, axis=0).T #np.tile(ss,nx).reshape(ny,nx)
    
    return sag,np.arange(nx),np.arange(ny)


    
def make_range(nx,ny):
    """make a data,x,y with increasing integer values for each pixel.
    
    example:

    >>> make_range(2,3)
    
    (array([[0., 1.],
        [2., 3.],
        [4., 5.]]), array([0, 1]), array([0, 1, 2]))"""
        
    data,x,y=np.arange(ny*nx,dtype=float).reshape(ny,nx),np.arange(nx),np.arange(ny)
    x=np.arange(nx)
    y=np.arange(ny)
    return data,x,y
    
def make_prof_legendre(x,coeff,inanp=[]):      
    """create a profile on x with legendre coefficients coeff.
    Add nans on inanp indices."""
    yp=np.polynomial.legendre.legval(x,coeff)
    yp[inanp]=np.nan
    return yp

def make_surf_legendre(x,y,coeff,inanp=[],inanl=[]):
    """Create a test surface from profiles along x, created with
    legendre polynomial with coefficients 'coeff'. 
    """
       
    data=np.empty((len(y),len(x)))
    nrep=len(x)//len(coeff)
    for i,cc in enumerate(coeff):
        yp=make_prof_legendre(y,cc,)
        data[:,i*nrep:(i+1)*nrep]=yp[:,None]
    data[inanp,inanp]=np.nan
    data[:,inanl]=np.nan

    return data

def make_surf_plane(x,y,coeff):
    """Create a test surface from plane coefficients `coeff` as returned from `pySurf.plane_fit`.
    based on
        # Ax + By + C = z    
    """
    
    A,B,C = coeff
    xx,yy = np.meshgrid(x,y)
    z = A * xx + B * yy + C
    data = z.reshape(len(y),len(x))

    return data
    
    
def test_profile_legendre(nanpoints=0):
    """test my routines with profile
    """
    
    x=np.arange(12)
    y=np.arange(20)
    coeff=[[0,3,-2],[12,0.01,-0.5],[-1,0.5,0]]
    inanp=[4,7,10] # indices for point nans if nanpoints != 0
    inanl=[2,5]     #indices for nan lines 

    plt.figure("fit and reconstruction")
    plt.clf()
    for cc in coeff:
        yp=np.polynomial.legendre.legval(y,cc)
        if nanpoints>0:
            yp[inanp]=np.nan
        plt.plot(y,yp,label='build %5.4f %5.4f %5.4f'%tuple(cc))
        yrec=fitlegendre(y,yp,2)
        plt.plot(y,yrec,'x')
    plt.legend(loc=0)

def test_profile_legendre(nans=True,fixnans=True):
    """test on a list of 2D coefficients creating profile then fitting with 
        numpy routines and with data2D routine fitlegendre 
        (wrapper around numpy to handle nans)."""
    
    y=np.arange(20)
    coeff=[[0,3,-2],[12,0.01,-0.5],[-1,0.5,0]]
    if nans:
        inanp=[4,7,10] # indices for point nans if nanpoints != 0
    else:
        inanp=[]
        
    plt.figure("fit and reconstruction")
    plt.clf()
    testdata=[y]
    for cc in coeff:
        yp=make_prof_legendre(y,cc,inanp)
        testdata=testdata+[yp]
        plt.plot(y,yp,label='build %5.4f %5.4f %5.4f'%tuple(cc))
        ccfit=np.polynomial.legendre.legfit(y,yp,2)
        print('fit:%5.4f %5.4f %5.4f'%tuple(ccfit))
        yrec=np.polynomial.legendre.legval(y,ccfit)
        yrec2=fitlegendre(y,yp,2,fixnans=fixnans)
        plt.plot(y,yrec,'x',label='rec  %5.4f %5.4f %5.4f'%tuple(ccfit))
        plt.plot(y,yrec2,'o',label='data2D.fitlegendre')
    plt.legend(loc=0)
    
    return testdata
    
def test_surf_legendre(nans=True,fixnans=True):
    """test how 1D (line) routines in polynomial.legendre work on 1D and 2D data.
    If nanpoints=0 nan are not put in data. If nanpoints=1 or 2 nans are added on some points and
    some lines, with value value of nanpoints determineing the option nanstrict of levellegendre
    (1=false, 2=true).
    """
    x=np.arange(12)
    y=np.arange(20)
    coeff=[[0,3,-2],[12,0.01,-0.5],[-1,0.5,0]]
    if nans:
        inanp=[4,7,10] # indices for point nans if nanpoints != 0
        inanl=[2,5]     #indices for nan lines    x=np.arange(12)
    else:
        inanp,inanl=[]
    
    data=make_surf_legendre(x,y,coeff,inanp,inanl)
    
    #test on surface
    ccfit=np.polynomial.legendre.legfit(y,data,2)
    print("shape of data array: ",data.shape)
    print("shape of coefficient array:",ccfit.shape)
    #print 'fit:%5.4f %5.4f %5.4f'%tuple(ccfit)
    #xx,yy=np.meshgrid(x,y)
    
    datarec=np.polynomial.legendre.legval(y,ccfit).T
    #datarec=np.polynomial.legendre.legval2d(yy,xx,ccfit)

    plt.figure("2D fit and reconstruction with numpy")
    plt.clf()    
    plt.subplot(131)
    plt.title('data')
    plt.imshow(data,origin='lower',interpolation='none')
    plt.colorbar()
    plt.subplot(132)
    plt.title('reconstruction')
    plt.imshow(datarec,origin='lower',interpolation='none')
    plt.colorbar()
    plt.subplot(133)
    plt.title('difference')
    plt.imshow(data-datarec,origin='lower',interpolation='none')
    plt.colorbar()
    plt.tight_layout()
    
    datarec=fitlegendre(y,data,2,fixnans=fixnans)
    plt.figure("2D fit and reconstruction with fitlegendre")
    plt.clf()    
    plt.subplot(131)
    plt.title('data')
    plt.imshow(data,origin='lower',interpolation='none')
    plt.colorbar()
    plt.subplot(132)
    plt.title('reconstruction')
    plt.imshow(datarec,origin='lower',interpolation='none')
    plt.colorbar()
    plt.subplot(133)
    plt.title('difference')
    plt.imshow(data-datarec,origin='lower',interpolation='none')
    plt.colorbar()
    plt.tight_layout()
    
    #test my routine now
    plt.figure("legendre removal with levellegendre")
    plt.imshow(levellegendre(y,data,2),origin='lower',interpolation='none')
    plt.colorbar()
    plt.show()
    
    #compare with old remove legendre
    plt.figure("compare removal routines")
    plt.clf()
    
    data=data+np.random.random(data.shape)
    datarec=levellegendre(y,data,2)
    rem4=removelegendre(y,2)
    rldata=level_by_line(data,rem4)
    
    plt.subplot(131)
    plt.imshow(datarec,origin='lower',interpolation='none')
    plt.title('levellegendre')
    plt.colorbar()
    plt.subplot(132)
    plt.imshow(rldata,origin='lower',interpolation='none')
    plt.title('level_by_line')
    plt.colorbar()
    plt.subplot(133)
    plt.imshow(datarec-rldata,origin='lower',interpolation='none')
    plt.title('difference')
    plt.colorbar()
    plt.tight_layout()

def test_makeGaussian(N=100,rnd=3.):
    a=makeGaussian(N)*20
    a=a+np.random.random((N,N))*rnd
    fig=plt.figure()
    fig.canvas.set_window_title('Gaussian surface N=%i, noise=%f'%(N,rnd))
    title="min:%4.1f max:%4.1f rms:%4.3f"%(np.nanmin(a),np.nanmax(a),np.nanstd(a))
    plt.title(title)
    plt.imshow(a,interpolation='none')
    plt.colorbar()
    plt.show()
    return a
    
if __name__=="__main__":
    N=100
    data=test_makeGaussian(100)
    print(data)




def power_law(x,k_n,N):
    """return a power law for points at x, 
    given scale factor k_n and exponent N as given by k_n*x**N """
    return k_n*x**N    
    
def fit_power_law(x,y,range=None):
    """fit a PSD according to a model y=K_N*x**N
    where x is in mm-1 and y in mm*3"""
    logx = np.log10(x)
    logy = np.log10(y)
    
    if not(range is None):
        i = np.where((logx > range[0]) & (logx < range[1]))
        logx = logx[i]
        logy = logy[i]
    out = np.polyfit(logx, logy, 1)
      
    return (10**out[1],out[0])  #offset, slope -> ampl., index





"""contains operations on dictionaries"""
import warnings
import inspect
import os



filterByKey2 = lambda data,keys : {key: data[key] for key in keys if key in data}

def pop_kw(kws,defaults=None,keylist=None,exclude=None):
    """A multiple pop function for a dictionary, hopefully useful to filter kws and manage function **kwargs parameters.

    Return a dictionary of values corresponding to kws[k] for k in keylist by popping
      the element from kws.
    If k is present in a dictionary `defaults`, take the values from `defaults` if not present in kws.
      return a dictionary {key:val} with val from `kws` or `defaults` if in any of them.
    Another way of seeing the return value is as a copy of `defaults` where values are replaced
      (after popping them) from same key in kws.


        kwargs={'dog':'bau','cat':'miao'}
        defs={'dog':'barf','sheep':'bee'}

        #pop 'cat' and 'sheep' if they are in kwargs or take them from defaults if there,
        #  put them in res
        >>> res=pop_kw(kwargs,defaults=defs)
        >>> kwargs
        Out[103]: {'cat': 'miao'}
        >>> res
        Out[104]: {'dog': 'barf', 'sheep': 'bee'}

        #pop 'cat' and 'sheep' if they are in kwargs or take them from defaults if there,
        #  put them in res
        >>> res=pop_kw(kwargs, defaults = defs, keylist = ['cat','sheep'])
        >>> kwargs
        Out[103]: {}
        >>> res
        Out[104]: {'cat': 'miao', 'dog': 'bau', 'sheep': 'bee'}


    A typical usage can be dividing arbitrary **kwargs between different subroutines,
      where the subroutines might not accept **kwargs, and be constrained to specific list of parameters.
      To pass kwargs in this circumstance do inside the routine:

    Module `inspect` can be used to retrieve subfunction parameter names and make a safe call, this is done automatically in `strip_kw`, equivalently to:

       import inspect
       pars=list(inspect.signature(f1).parameters) #give list of all accepted parameters in f1
       res1=f1(arg1,**pop_kw(kwargs,keylist = pars)) #note that pars will have keys 'args' and 'kwargs' if
                            # f1 accepts arbitrary arguments. kws and defaults with same name can be
                            #  in conflict

    In general, if **kws are passed to a function that calls two other subs, each one with optional arguments, it can be used as (see full example in `dataIO.dicts.example_pop_kw`):

        def example_pop_kw(**kwargs):

            # here if passed directly with `res0=f1(arg1,**kwargs)` gives error because mes2 is not a defined parameter
            res1=f1(arg1,**pop_kw(kwargs,{'mes1':'blah!'},['mes1'])) #only mes1 is passed to res1 with default if not in kwargs:
            res2=f2(arg1,**kwargs)  #other remaining kwargs can be safely passed to f2

    2020/05/26 added `exclude` keywork, as a consequence of adding to strip_kw.
    Completely reviewed 2019/04/08, changed interface and function of keylist.


    """
    res={}
    keys=list(defaults.keys())
    
    if exclude is None: exclude = [] #makes a list, so I can use `in`
    for k in keys:
        #print("%s %s %s %s"%(k,defaults[k],res,kws))
        if k in exclude:
            pass
            #print ('Excluded')
        else:
            res[k]=kws.pop(k,defaults[k])
    #print("final: %s %s"%(res,kws))
    
    if keylist is not None:
        if isinstance(keylist,str): keylist=[keylist] #make list if only one element
        for k in keylist:
            if k in kws:
                res[k]=kws.pop(k)

    return res

    '''
	def pop_kw(kws,defaults,keylist=None):
    """A multiple pop function for a dictionary, hopefully useful to filter kws and manage function parameters.

    a copy of `defaults` is returned where values are stripped from kws if a key is present.
    If **kws are passed to a function that calls two other subs, each one with optional arguments, it can be used as (see full example below):
    res1=sub1(arg1,**pop_kw(kws,{'mes1':'blah!'}))
    res2=sub2(arg1,**kws)

    or values can be retrieved in variables (and sorted):
    mes2,mes1=pop_kw(kws,{'mes1':'blah!'},['mes2','mes1'])
    print (mes1,mes2)

    Note however that only parameters explicitly listed in defaults are filtered. A safer version for filtering function parameters rather than generic dictionaries is strip_kw.

    """
    res={}
    keys=list(defaults.keys())
    for k in keys:
        res[k]=kws.pop(k,defaults[k])

    if keylist is not None:
        try:
            res=[res[k] for k in keylist]
        except KeyError as e:
            print ("ignoring key invalid in dict %s:\n'"%(res))
            warnings.warn(str(e),RuntimeWarning)

    return res
	'''

def strip_kw(kws,funclist,split=False,exclude=None,**kwargs):
    """ Enhanced version of pop_kw, kw can be extracted from inspection of a function (or a list  of functions).
    Defaults can be passed as kwargs.
    `exclude` is a list of keys that are kept in kws unchanged
       and not returned.
    
    2020/05/26 added `exclude`.
    Use case is presented from problems with psd_analysis.
    `psd2d.psd_analysis` calls `strip_kw(kwargs,psd2d_analysis,title="")`. 
    `ps2d_analysis` has keyword `units`, that is not used in this case (used, only if called with `analysis=True`). Here we want to preserve `units` to strip it later.
    
    Old notes:
        was in theory a safer version of the first pop_kw.
           I am not sure in what it was supposed to be "safer",
       but it was accepting functions as input.
	   It might have been used a negligible amount of times.

       sub_kw=strip_kw(kws,[sub],def1=val1,def2=val2)"""

    res=[]
    if callable(funclist): #if scalar, makes funclist a one element list
        funclist=[funclist]    
        
    for func in funclist:
        f=inspect.getfullargspec(func)
        defaults=kwargs
        l = len(f.defaults) if f.defaults is not None else 0
        fdic=dict(zip(f.args[-l:],f.defaults))
        res.append(pop_kw(kws,fdic,exclude=exclude))

    if not split:  #merge dictionaries, don't ask me how
        tmp = {}
        [tmp.update(d) for d in res]
        res=tmp

    return res

def print_tree(tree, depth = 0):
    """print dictionary as tree, recursively calling itself,
    based on http://www.siafoo.net/snippet/91"""
    if tree == None or len(tree) == 0:
        #print () #"--" * depth)
        pass
    else:
        for key, val in tree.items():
            #print ("\t" * depth, key)
            #printTree(lst_view_path(val), depth+1)    
            
            
            if isinstance(val,dict): #folder:
                print ("\t" * depth+'|', key.upper(),"----\\")  #print ("\t" * depth, key.upper())
                print_tree(val, depth+1)    
            else:  #file:
                print ("\t" * depth+'|', key,':',val[:6])  #print ("\t" * depth, key,':',val[:6])
                # non so perche' qui sopra si limita a 6, forse per array grandi, ma fallisce se altri dati.
               
 
def test_filterByKey():
	eegKeys = ["FP3", "FP4"]
	gyroKeys = ["X", "Y"]

	# 'Foo' is ignored
	data = {"FP3": 1, "FP4": 2, "X": 3, "Y": 4, "Foo": 5}
	print(filterByKey(data,eegKeys))

def example_pop_kw():

    def f1(arg,mes1='--'):
        print (arg,mes1)
    def f2(arg,mes2='--'):
        print (arg,mes2)

    def function(arg1,**kwargs):
        '''a function with 1 arg. pop_kw is used to manipulate and smartly filter `**kwargs` between two functions,
          `f1(arg,mes1='--')` and `f2(arg,mes2='--')`, so not to pass unexpected arguments.
          `res0=f1(arg1,**kwargs)` #would give error.'''

        print ("arguments passed to functions: \n arg:",arg1,"\nkwargs:",kwargs)

        # here if passed directly with `res0=f1(arg1,**kwargs)` because mes2 is not a defined parameter
        res1=f1(arg1,**pop_kw(kwargs,'mes1',{'mes1':'blah!'})) #only mes1 is passed to res1 with default if not in kwargs:

        #other remaining kwargs can be passed to f2
        res2=f2(arg1,**kwargs)

        return res1,res2


    function('This is:',mes2='mes2!')
    print('**')
    function('This is:',mes1='mes1!')
    print('**')
    function('This is:',mes1='mes1!',mes2='mese2!')
    print('**')
    function('This is:',mes1='mes1!',mes2='mese2!')

#from dataIO.dicts import pop_kw, strip_kw
def test_pop_kw(function=pop_kw):#function = None):

    """run a set of dict stripping tests using a function with interface compatible with `pop_kw`.
    uses pop_kw if function is not provided."""

    from inspect import signature
    
    print (signature(function))
    print (function.__doc__)

    print("Without keylist return a dictionary using only keys in defaults, used keys are removed from kwargs:\n")
    kwargs={'dog':'bau','cat':'miao','cow':'muu'}
    defs={'dog':'barf','sheep':'bee'}
    print('start -> kwargs = {},\n defaults = {}'.format(kwargs,defs))
    res=function(kwargs,defaults=defs)  #remove from kwargs the keys in defs assigning default if they don't exist in kwargs
    print("\nResult: ",res)
    # {'sheep': 'bee', 'dog': 'bau'}
    print('end ->\n kwargs = {},\n defaults = {}'.format(kwargs,defs))
    print("#---------------")

    print("Keys to keep can be specified separately passing keylist:\n")
    kwargs={'dog':'bau','cat':'miao','cow':'muu'}
    defs={'dog':'barf','sheep':'bee'}
    keys=['cat','sheep']
    print('start -> kwargs = {},\n keylist = {},\n defaults = {}\n'.format(kwargs,keys,defs))
    res=function(kwargs,defaults=defs,keylist=keys)
    print("\nResult: ",res)
    # {'sheep': 'bee', 'cat': 'miao', 'dog': 'bau'}
    print('end ->\n kwargs = {},\n keylist = {},\n defaults = {}'.format(kwargs,keys,defs))
    print("#---------------\n")

    print("Same, with added keys, when not present in kwargs, they are ignored:\n")

    kwargs={'dog':'bau','cat':'miao','cow':'muu'}
    defs={'dog':'barf','sheep':'bee'}
    keys = ['nsigma','units','gatto']
    print('start ->\n kwargs = {},\n keylist = {},\n defaults = {}\n'.format(kwargs,keys,defs))
    res=function(kwargs,defs,keys) #gatto doesn't exist in kwargs
    print("\nResult: ",res)
    # ignoring key invalid in dict {'units': 'mm', 'nsigma': 1}:
    #'
    #{'sheep': 'bee', 'dog': 'bau'}
    print('end ->\n kwargs = {},\n keylist = {},\n defaults = {}'.format(kwargs,keys,defs))
    print("#------------")

    print("Use `exclude`, keys in this list are left untouched in kwargs.\n")

    kwargs={'dog':'bau','cat':'miao','cow':'muu'}
    defs={'dog':'barf','sheep':'bee'}
    exclude = ['dog']
    print('start ->\n kwargs = {},\n exclude = {},\n defaults = {}\n'.format(kwargs,exclude,defs))
    res=function(kwargs,defs,exclude=exclude) #gatto doesn't exist in kwargs
    print("\nResult: ",res)
    # res =  {'sheep': 'bee'} #dog is not modified or popped
    # kwargs = {'dog': 'bau', 'cat': 'miao', 'cow': 'muu'}
    print('end ->\n kwargs = {},\n exclude = {},\n defaults = {}'.format(kwargs,exclude,defs))
    print("#------------")

if __name__=="__main__":
    test_pop_kw()





def chk_args_id(func,n=None,*args,**kwargs):
    """check the first n arguments id and print a warning if
    they don't match the first n return arguments."""

    def wrapper(*args,**kwargs):

        ids = [id(a) for a in args]

        res=func(*args,**kwargs)
        
        ids2 = [id(a) for a in res]
        
        if ids != ids2:
            print("ids not matching.")
        else: 
            print("ids matching")
        print ("ids (old,new):\n",ids,ids2)
    return wrapper
    
    
def test_func(x,y,z,replacez=False):
    if replacez: z='replaced' #this should make the test fail
     
    return x,y,z
    
def test_decorator():
    
    print (test_func(1,2,3))
    t=chk_args_id(test_func)
    t(1,2,3)
    
    print (test_func(1,2,3,True))
    t=chk_args_id(test_func)
    t(1,2,3,True)
    
    """
    a,b,c=[1,2],-6,['cane']
    print (test_func(a,b,c))
    t=chk_args_id(test_func)
    t(a,b,c)
    
    print (test_func(a,b,c,True))
    t=chk_args_id(test_func)
    t(a,b,c,True)
    """

def decorate_module(mod,dec):
    """decorate all functions in a given module passed as string. it probably will not work, how export to caller routines?

    modified from:  https://stackoverflow.com/questions/8718885/import-module-from-string-variable?utm_medium=organic&utm_source=google_rich_qa&utm_campaign=google_rich_qa
    """
    
    import importlib

    i = importlib.import_module(mod)

    import types
    for k,v in vars(i).items():
        if isinstance(v, types.FunctionType):
            vars(i)[k] = dec(v)    
    
if __name__=="__main__":
    test_decorator()


"""
2021/06/15 header is now implemented in data2D.read_data, it calls the function in format_reader with header flag, however there is not exception handling, so I am not sure what happens when the flag is not supported.

2020/07/14 see notes on 02/29. raw readers in format_reader should
   return data with the minimal amount of manipulation and the most general interface. All format/instrument specific aspects
   should be handled there.
   TODO: At the moment `header` flag (set to return header rather than data), is handled by `data2D.read_data` routine called by readers in this file, that wraps the raw reader,
   and registers data. `Data2D.__init__` method when called with a file also fill the property `.header`, this should be handled by calling read_data with     
   , needs to be handled here. Note that some raw readers retrieve header information independently (e.g. as property `header` of an object. In this case, all the handling is done here.
   
   Note also that the entire module might be unnecessary, many of the reader functions say: '''temporary wrapper for new readers, replace call to matrixZygo_reader
    with calls to read_data(wfile,reader=csvZygo_reader)'''

2020/05/13 change import calls to use pySurf.readers.format_reader instead of _instrument_reader

2020/03/17 

2020/02/29 sto definendo meglio la struttura.
_instrument_reader contiene i raw reader, che restituiscono data,x,y come
letti da file con le informazioni necessarie. Questi hanno argomento header, che permette di restituire l'header.

Le funzioni contenute in questo modo devono permettere di restituire funzioni Reader derivate da raw_reader applicando read_data(che esegue registrazioni e allineamento posteriori ai raw e come preprocessing).
Una funzione reader Reader fissa alcuni argomenti per un raw reader e/o per read_data (che devono poi essere overridable alla chiamata).

---------------


2018/06/05 v3 the most used reading functions are replaced by a wrapper around read_data that provides a temporary interface.
read_data distributes passed parameters to the reader and to the data registration (data2D.register_data) providing a temporary interface.
Removed original code for functions.

A reader function in this module is a function, specific to a file format, that accepts a filename, a set of format-specific parameters, a set of registering parameters (register_data), and return properly registered data,x,y.

It is implemented here as a wrapper around read_data with the argument reader set to the proper raw reader (from _instrument_reader).
read_data is needed to divide parameters to be pased to the raw_reader, while the fact that here we are defining each reader is just for interface consistency.
e.g.:

    data,x,y = matrixZygo_reader(file,*args,**kwargs)

is equivalent to:

    data,x,y = read_data(file,reader=csvZygo_reader,*args,**kwargs)

Or could be defined as partial functions or closures.

A call to Omitting register_data parameters has the effect of skipping registration.
A dictionary with supported extension and associated readers is

2018/05/08 v1.2 with common interface for data preprocessing
(scaling, aligning, centering, etc.. ).

Start migration by writing wrapper functions with old names, calling core functions in _instrument_reader.
Once interface is designed"""


import numpy as np
from pySurf.data2D import crop_data, remove_nan_frame, register_data
from astropy.io import fits
from pySurf.points import get_points
from pySurf.points import crop_points
from pySurf.points import points_find_grid
from pySurf.points import resample_grid
from .read_sur_files import readsur
from pySurf.data2D import read_data

#from utilities.imaging.man import stripnans
from dataIO.read_pars_from_namelist import read_pars_from_namelist

from pySurf.points import matrix_to_points2
import pdb

#from pySurf.readers._instrument_reader import *
#from pySurf.readers._instrument_reader import read_data
#from pySurf.readers._instrument_reader import reader_dic
#from pySurf.readers._instrument_reader import points_reader
'''
2020/07/13 removed, use alt version in data2D
def read_data(file,reader,register=True,*args,**kwargs):
    """non essendo sicuro dell'interfaccia per ora faccio cosi'.
    The function first calls the (raw) data reader, then applies the register_data function to address changes of scale etc,
    arguments are filtered and passed each one to the proper routine.
    18/06/18 add all optional parameters, if reader is not passed,
    only registering is done. note that already if no register_data
    arguments are passed, registration is skipped.
    18/06/18 add action argument. Can be  'read', 'register' or 'all' (default, =read and register). This is useful to give fine control, for example to modify x and y after reading and still make it possible to register data (e.g. this is done in
    Data2D.__init__).

    implementation:
    This works well, however read_data must filter the keywords for the reader and for the register and
    this is hard coded, that is neither too elegant or maintainable. Note however that with this structure it is
    possible to call the read_data procedure with specific parameters, for example in example below, the reader for
    Zygo cannot be called directly with intensity keyword set to True without making a specific case from the other readers,
    while this can be done using read_data. """

    """this is an ugly way to deal with the fact that return
    arguments are different if header is set, so when assigned to a variable as in patch routines in pySurf instrumentReader it fails.
    Workaround has been calling directly read_data, not optimal."""
    if kwargs.pop('header',False):
        try:
            h = reader(file,header=True)
            return h if h else ""
        except TypeError:  #unexpected keyword if header is not implemented
            return None

    #filters register_data parameters cleaning args
    # done manually, there is a function in dataIO.
    scale=kwargs.pop('scale',(1,1,1))
    crop=kwargs.pop('crop',None)
    center=kwargs.pop('center',None)
    strip=kwargs.pop('strip',False)
    regdic={'scale':scale,'crop':crop,'center':center,'strip':strip}

    #pdb.set_trace()
    data,x,y=reader(file,*args,**kwargs)

    #try to modify kwargs
    for k in list(kwargs.keys()): kwargs.pop(k)  #make it empty
    #kwargs.update(regdic)
    for k in regdic.keys():kwargs[k]=regdic[k]
    if register:
        data,x,y=register_data(data,x,y,scale=scale,crop=crop,
        center=center,strip=strip)
    return data,x,y
    '''

def xread_data(file,reader,register=True,*args,**kwargs):
    """non essendo sicuro dell'interfaccia per ora faccio cosi'.
    The function first calls the (raw) data reader, then applies the register_data function to address changes of scale etc,
    arguments are filtered and passed each one to the proper routine.
    18/06/18 add all optional parameters, if reader is not passed,
    only registering is done. note that already if no register_data
    arguments are passed, registration is skipped.
    18/06/18 add action argument. Can be  'read', 'register' or 'all' (default, =read and register). This is useful to give fine control, for example to modify x and y after reading and still make it possible to register data (e.g. this is done in
    Data2D.__init__).

    implementation:
    This works well, however read_data must filter the keywords for the reader and for the register and
    this is hard coded, that is neither too elegant or maintainable. Note however that with this structure it is
    possible to call the read_data procedure with specific parameters, for example in example below, the reader for
    Zygo cannot be called directly with intensity keyword set to True without making a specific case from the other readers,
    while this can be done using read_data. """

    """this is an ugly way to deal with the fact that return
    arguments are different if header is set, so when assigned to a variable as in patch routines in pySurf instrumentReader it fails.
    Workaround has been calling directly read_data, not optimal."""
    if kwargs.pop('header',False):
        try:
            return reader(file,header=True)
        except TypeError:  #unexpected keyword if header is not implemented
            return None


    #filters register_data parameters cleaning args
    # done manually, there is a function in dataIO.
    scale=kwargs.pop('scale',(1,1,1))
    crop=kwargs.pop('crop',None)
    center=kwargs.pop('center',None)
    strip=kwargs.pop('strip',False)
    regdic={'scale':scale,'crop':crop,'center':center,'strip':strip}

    data,x,y=reader(file,*args,**kwargs)

    #try to modify kwargs
    for k in list(kwargs.keys()): kwargs.pop(k)  #make it empty
    #kwargs.update(regdic)
    for k in regdic.keys():kwargs[k]=regdic[k]
    if register:
        data,x,y=register_data(data,x,y,scale=scale,crop=crop,
        center=center,strip=strip)
    return data,x,y

def matrixZygo_reader(wfile,*args,**kwargs):
    """temporary wrapper for new readers, replace call to matrixZygo_reader
    with calls to read_data(wfile,reader=csvZygo_reader)"""
    from pySurf.readers.format_reader import csvZygo_reader
    #pdb.set_trace()
    return read_data(wfile,csvZygo_reader,*args,**kwargs)

def matrix_reader(wfile,*args,**kwargs):
    """temporary wrapper for new readers, replace call to matrix4D_reader
    with calls to read_data(wfile,reader=matrix4D_reader)"""
    from pySurf.data2D import data_from_txt
    return read_data(wfile,data_from_txt,*args,**kwargs)

def matrix4D_reader(wfile,header=False,delimiter=',',*args,**kwargs):
    """temporary wrapper for new readers, replace call to matrix4D_reader
    with calls to read_data(wfile,reader=matrix4D_reader).
    
    read csv data with parameters extracted from header as saved by 4D.
    
    """
    from pySurf.readers.format_reader import csv4D_reader

    #import pdb
    #pdb.set_trace()
    return read_data(wfile,csv4D_reader,delimiter=delimiter,*args,**kwargs)


def matrixsur_reader(wfile,*args,**kwargs):
    """temporary wrapper for new readers, replace call to matrixsur_reader
    with calls to read_data(wfile,reader=sur_reader)"""
    from pySurf.readers.format_reader import sur_reader
    return read_data(wfile,sur_reader,*args,**kwargs)

def matrixdat_reader(wfile,*args,**kwargs):
    """temporary wrapper for zygo metropro .dat binary files"""
    from pySurf.readers.format_reader import datzygo_reader
    return read_data(wfile,datzygo_reader,*args,**kwargs)

def nid_reader(wfile,*args,**kwargs):
    """temporary wrapper for new readers,
    with calls to read_data(wfile,reader=fits_reader)"""
    from pySurf.readers.format_reader import read_nid
    return read_data(wfile,read_nid,*args,**kwargs)
    
def points_reader(wfile,*args,**kwargs):
    """temporary wrapper for points readers, read and register
    points and convert to 2D data"""
    from pySurf.readers.format_reader import points_reader
    return read_data(wfile,points_reader,*args,**kwargs)

### UNWRAPPED FUNCTIONS:
def FEAreader(FEAfile):
    """read V. Marquez files, keeping original orientation and variables.
    """
    xx,yy,zz,dx,dy,dz=np.genfromtxt(FEAfile,delimiter=',',skip_header=1,
        unpack=1,usecols=(2,3,4,5,6,7))
    #a=np.genfromtxt(FEAfile,delimiter=',',skip_header=1)
    #xx,yy,zz,dx,dy,dz=np.hsplit(a[:,2:8],6)
    x=-xx-dx
    y=yy+dy  #Correct for gravity
    z=dz  #here I keep only the deformation, and ignore nominal cylinder shape
    p=np.vstack([x,y,dz]).T*1000. #convert to mm, all coordinates
    return p

def getdata(fitsfile):
    """return x,y (vectors) and data (matrix) from fits file."""
    #works with fits from CCI sur files
    data,x,y=fits_reader(fitsfile)
    header=fits_reader(fitsfile,header=True)

    #from pySurf.points import level_points
    #data=level_points(data)

    dx=header['DX']
    dy=header['DY']
    x=dx*np.arange(data.shape[0])/1000. #convert in mm
    y=dy*np.arange(data.shape[0])/1000.

    return x,y,data


def fits_reader(wfile,*args,**kwargs):
    """temporary wrapper for new readers, replace call to fits_reader
    with calls to read_data(wfile,reader=fits_reader)"""
    from pySurf.readers.format_reader import fits_reader
    return read_data(wfile,fits_reader,*args,**kwargs)
    
def fitsWFS_reader(wfile,ypix=1.,ytox=1.,zscale=1.,crop=[None,None,None],
    center=None,strip=False,scale=(1,1,1.)):
    """return data (matrix) and x,y (vectors) from fits WFS file. Here all lateral sizes must be speficied or default (no header info).
    I don't  understand very well order of data, inverting data
    works for S22 for bump positive, but sample seems rotated and
    horizontally flipped.
    If strip is set, nan are removed. strip is done after centering and scaling, so center is calculated on full data."""
    
    from pySurf.readers.format_reader import fits_reader
    
    data,x,y=fits_reader(fitsfile)

    #adjust crop and scales
    ny,nx=data.shape
    x=np.arange(nx)*ypix*ytox*nx/(nx-1)*scale[0]
    y=np.arange(ny)*ypix*ny/(ny-1)*scale[1]
    data=data*zscale*scale[2]

    #if any x or y is inverted invert data and orient as cartesian.
    if x[-1]<x[0]:
        x=x[::-1]
        data=np.fliplr(data)
        x=x-min(x)
    if y[-1]<y[0]:
        y=y[::-1]
        data=np.flipud(data)
        y=y-min(y)

    #doing cropping here has the effect of cropping on cartesian orientation,
    #coordinates for crop are independent on center and dependent on scale and pixel size.
    #center is doing later, resulting in center of cropped data only is placed in the suitable position.
    if crop is None:
        crop=[None,None,None]
    data,x,y=crop_data(data,x,y,*crop)

    #center data if center is None, leave unchanged
    if center is not None:
        assert len(center)>=2
        x=x-(np.max(x)+np.min(x))/2.+center[0]
        y=y-(np.max(y)+np.min(y))/2.+center[1]
        if len(center)==3:
            data=data-(np.nanmax(data)+np.nanmin(data))/2.+center[2]

    if strip:  #messes up x and y
        data,x,y = remove_nan_frame(data,x,y)

    return data,x,y

def fitsAFM_reader(fitsfile,sizeum=None):
    """return data, x,y from fits file. accept size um to calculate x and y.AFM file"""
    
    from pySurf.readers.format_reader import fits_reader
    data,x,y=fits_reader(fitsfile)
    #z in m
    from pySurf.points import level_points
    data=level_points(data)

    assert data.shape[0]==data.shape[1]  #if not, it's first time it happens, check axis.
    x=np.linspace(1,sizeum[0],data.shape[0])/1000. #convert in mm
    y=np.linspace(1,sizeum[1],data.shape[1])/1000.
    data=data*1e6  #convert from m to um
    return matrix_to_points2(data,x,y)

def fitsCCI_reader(fitsfile,center=None,header=False):
    """return x,y (vectors) and data (matrix) from fits CCI file. `x` and `y` are obtained from 'DX' and 'DY' keys in header. 
    center can be none, in which case data are not centered.
    To extract center position from header, use center=(None,None)."""
    
    from pySurf.readers.format_reader import fits_reader
    
    data,x,y=fits_reader(fitsfile)
    HH=fits_reader(fitsfile,header=True)

    dx=HH['DX']
    dy=HH['DY']
    x=dx*np.arange(data.shape[0])/HH['XUNITRAT'] #convert in mm
    y=dy*np.arange(data.shape[1])/HH['YUNITRAT']

    if center is not None:
        assert len(center)>=2
        if (center is None).any():
            center=(HH['XOFFSET'],HH['YOFFSET'])
        x=x-(np.max(x)+np.min(x))/2.+center[0]
        y=y-(np.max(y)+np.min(y))/2.+center[1]
        if len(center)==3:
            data=data-(np.nanmax(data)+np.nanmin(data))/2.+center[2]

    if header:
        return HH
    else:
        return data,x,y


def bin_reader(wfile,index=0,ytox=1.,zscale=1.,crop=[None,None,None],center=None):
    """Read a binary .npy file containing a list of lists in form (imdata, dx).
    Can crop data and translate on three coordinates. crop is calculated on raw data,
    center is the position of the central point after crop and scaling."""

    d=np.load(wfile)
    data=d[index][0]
    ypix=d[index][1]
    ny,nx=data.shape
    x=np.arange(nx)*ypix*ytox*nx/(nx-1)
    y=np.arange(ny)*ypix*ny/(ny-1)
    y=max(y)-y
    data=data*zscale
    data,x,y=crop_data(data,x,y,*crop)
    if center is not None:
        assert len(center)>=2
        x=x-(np.max(x)+np.min(x))/2.+center[0]
        y=y-(np.max(y)+np.min(y))/2.+center[1]
        if len(center)==3:
            data=data-(np.nanmax(data)+np.nanmin(data))/2.+center[2]

    return data,x,y

def test_datzygo_reader (wfile=None):
    
    import os
    import matplotlib.pyplot as plt
    from  pySurf.data2D import plot_data
    from pySurf.data2D_class import Data2D
    from pySurf.readers.format_reader import datzygo_reader

    if wfile is  None:
        #relpath
        #wfile=r'/Volumes/GoogleDrive/Il mio Drive/progetti/sandwich/dati/09_02_27/botta di culo 1a.dat'
        wfile=r'C:\Users\kovor\Documents\python\pyXTel\source\pySurf\test\input_data\readers\botta di culo 1a.dat'
        
        #wfile= os.path.join(os.path.dirname(__file__),relpath)
    (d1,x1,y1)=datzygo_reader(wfile)
    dd1=Data2D(d1,x1,y1)
    
    dd2=Data2D(wfile,reader=datzygo_reader)
    
    plt.figure()
    plt.suptitle(os.path.basename(wfile))
    plt.subplot(121)
    dd1.plot(aspect='equal')
    plt.title('object from data')
    plt.subplot(122)
    dd2.plot(aspect='equal')
    plt.title('object from reader ')
    return dd2




if __name__=="__main__":

    def nonenizer(scale=None):

        if scale is None:
            scale=np.array((1.,1.,1.))
        else:
            if len(scale) == 2:
                scale=(scale[0],scale[1],1.)
            if len(scale) == 1:
                scale=(scale,scale,1.)
        return scale


    """test nonenizer"""
    a=5
    b=[5]
    c=[5,]
    d=np.array(a)
    e=np.array(b)
    f=np.array(c)

    for i in [a,b,c,d,e,f]:
        print(np.array(i).size)
